"""This module provides printing support.

Copyright 2014--2022 Michael Hayes, UCECE

"""

import re
from .config import latex_expr_map, pretty_expr_map, str_expr_map
from .config import functions, words, subscripts
from .latex import latex_str
from .state import state
from sympy.printing.str import StrPrinter
from sympy.printing.latex import LatexPrinter
from sympy.printing.pretty.pretty import PrettyPrinter
from sympy.printing.printer import print_function
import sympy as sym

__all__ = ('pretty', 'pprint', 'latex', 'print_str')

# Note, there are some magic hooks that are used for printing custom types:
#
# jupyter looks for methods called _repr_latex_ and _repr_pretty_.
#
# IPython looks for _repr_pretty
#
# sympy.latex() looks for methods called _latex

# LaTeX markup is nicer but it requires mathjax.

# FIXME: should import from parser
cpt_names = ('C', 'E', 'F', 'G', 'H', 'I', 'L',
             'R', 'NR', 'V', 'Y', 'Z', 'i', 'v')
cpt_name_pattern = re.compile(r"(%s)([\w']*)" % '|'.join(cpt_names))
sub_super_pattern = re.compile(r"([_\^]){([\w]+)}")
func_pattern = re.compile(r"\\operatorname{(.*)}")
word_name_pattern = re.compile(r"(%s)([\w']*)" % '|'.join(words))


def canonical_name(name):
    """Convert symbol name to canonical form for printing.

    R_{out} -> R_out
    R1 -> R_1
    XT2 -> XT_2
    Vbat -> V_bat
    alpha0 -> alpha_0
    """

    def foo(match):
        return match.group(1) + match.group(2)

    if not isinstance(name, str):
        return name

    if '\\' in name:
        return name

    # If the symbol name was created with symbol() or symbols()
    # then keep it as is.
    if state.context.symbols.kind(name) == 'user':
        return name

    # Convert R_{out} to R_out for SymPy to recognise.
    name = sub_super_pattern.sub(foo, name)

    # Convert v_Cfoo to v_C_foo, etc.   This is required for
    # state-space state variables.
    if (len(name) >= 3 and name[0:2] in ('v_', 'V_', 'i_', 'I_') and
            name[2] in ('L', 'C')):
        return name[0:2] + canonical_name(name[2:])

    if name.find('_') != -1:
        return name

    # Don't touch things like Heaviside
    if name.lower() in words + functions:
        return name

    # Convert R1 to R_1, etc.
    match = cpt_name_pattern.match(name)
    if match:
        if match.groups()[1] == '':
            return name
        name = match.groups()[0] + '_' + match.groups()[1]
        return name

    if len(name) < 2:
        return name

    # Convert omega1 to omega_1, etc.
    match = word_name_pattern.match(name)
    if match:
        if match.groups()[1] == '':
            return name
        name = match.groups()[0] + '_' + match.groups()[1]
        return name

    # Convert i1 to i_1, etc.
    if name[1].isdigit:
        return name[0] + '_' + name[1:]

    # Convert irms to i_rms, etc.
    if name[1:].lower() in subscripts:
        return name[0] + '_' + name[1:]

    return name


class LcapyStrPrinter(StrPrinter):

    def _print(self, expr, exp=None):

        from .equation import Equation
        from .expr import Expr

        if isinstance(expr, Equation):
            return self._print_Equation(expr)

        if isinstance(expr, Expr):
            expr = expr.expr

        # Convert sym.I to j etc.
        try:
            if expr in pretty_expr_map:
                return str_expr_map[expr]
        except:
            pass
        return super(LcapyStrPrinter, self)._print(expr)

    def _print_Equation(self, eq):

        return self._print(eq.lhs) + ',' + self._print(eq.rhs)

    def _print_Symbol(self, expr):
        # Do not canonicalise name since this is required for name
        # matching.  The only caller is the __str__ method for Expr.
        expr = sym.Symbol(expr.name)
        return super(LcapyStrPrinter, self)._print_Symbol(expr)


class LcapyLatexPrinter(LatexPrinter):

    def __init__(self, settings=None):
        LatexPrinter.__init__(self, settings)

    def _print(self, expr, exp=None):

        from .equation import Equation
        from .sequence import Sequence

        if isinstance(expr, Sequence):
            return self._print_Sequence(expr)
        elif isinstance(expr, Equation):
            return self._print_Equation(expr)

        if hasattr(expr, '_pexpr'):
            expr = expr._pexpr

        # Convert sym.I to j etc.
        try:
            if expr in latex_expr_map:
                return latex_expr_map[expr]
        except:
            pass

        if exp is None:
            return super(LcapyLatexPrinter, self)._print(expr)
        return super(LcapyLatexPrinter, self)._print(expr, exp=exp)

    def _print_Equation(self, eq):

        return self._print(eq.lhs) + ' = ' + self._print(eq.rhs)

    def _print_Sequence(self, seq):

        a = seq.zeroextend()

        items = []
        if seq.start_trunc:
            items.append('\\ldots')

        for v1, n1 in zip(a, a.n):
            try:
                s = v1.latex()
            except:
                s = str(v1)

            if n1 == 0:
                s = r'\underline{%s}' % v1
            items.append(s)

        if seq.end_trunc:
            items.append('\\ldots')

        return r'\left\{%s\right\}' % ', '.join(items)

    def _print_Piecewise(self, expr):

        if len(expr.args) > 1:
            return super(LcapyLatexPrinter, self)._print_Piecewise(expr)

        e, c = expr.args[0]
        return r"%s \;\; \text{for}\: %s" % (self._print(e), self._print(c))

    def _print_Heaviside(self, expr, exp=None):

        tex = r"%s\left(%s\right)" % (latex_expr_map[sym.Heaviside],
                                      self._print(expr.args[0]))
        if exp:
            tex = r"{%s}^{%s}" % (tex, exp)
        return tex

    def _print_UnitImpulse(self, expr, exp=None):

        tex = r"\delta\left[%s\right]" % self._print(expr.args[0])
        if exp:
            tex = r"{%s}^{%s}" % (tex, exp)
        return tex

    def _print_UnitStep(self, expr, exp=None):

        tex = r"u\left[%s\right]" % self._print(expr.args[0])
        if exp:
            tex = r"{%s}^{%s}" % (tex, exp)
        return tex

    def _print_dtrect(self, expr, exp=None):

        tex = r"\mathop{\mathrm{rect}}\left[%s\right]" % self._print(
            expr.args[0])
        if exp:
            tex = r"{%s}^{%s}" % (tex, exp)
        return tex

    def _print_dtsign(self, expr, exp=None):

        tex = r"\mathop{\mathrm{sign}}\left[%s\right]" % self._print(
            expr.args[0])
        if exp:
            tex = r"{%s}^{%s}" % (tex, exp)
        return tex

    def _print_symbol_name(self, name):

        name = canonical_name(name)

        parts = name.split('_')

        expr = sym.Symbol(parts[0])
        s = super(LcapyLatexPrinter, self)._print_Symbol(expr)

        if '_' in s:
            parts2 = s.split('_')
            s = parts[0]
            parts = parts2[1:] + parts[1:]

        if len(parts) == 1:
            return latex_str(s)

        if len(parts) == 2:
            return latex_str(s + '_{%s}' % parts[1])

        # Sympy cannot print a symbol name with a double subscript
        # using LaTeX.  This should be fixed in Sympy.
        # Need to convert v_C_1 to v_{C_{1}}

        if len(parts) == 3:
            return latex_str(s + '_{%s_{%s}}' % (parts[1], parts[2]))

        raise ValueError(
            'Cannot handle more than two subscripts for %s' % name)

    def _print_Symbol(self, expr):

        return self._print_symbol_name(expr.name)

    def _print_AppliedUndef(self, expr, exp=None):

        s = self._print_symbol_name(expr.func.__name__)
        args = [str(self._print(arg)) for arg in expr.args]
        tex = '%s(%s)' % (s, ','.join(args))
        if exp:
            tex = r"{%s}^{%s}" % (tex, exp)
        return tex


class LcapyPrettyPrinter(PrettyPrinter):

    def _print(self, expr, exp=None):

        from .sequence import Sequence
        from .equation import Equation

        if isinstance(expr, Sequence):
            return self._print_Sequence(expr)
        elif isinstance(expr, Equation):
            return self._print_Equation(expr)

        if hasattr(expr, '_pexpr'):
            expr = expr._pexpr

        try:
            if expr in pretty_expr_map:
                return self._print_basestring(pretty_expr_map[expr])
        except:
            pass
        return super(LcapyPrettyPrinter, self)._print(expr)

    def _print_Equation(self, eq):

        from sympy.printing.pretty.stringpict import prettyForm, stringPict

        op = prettyForm(' = ')

        l = self._print(eq.lhs)
        r = self._print(eq.rhs)
        pform = prettyForm(*stringPict.next(l, op, r), binding=prettyForm.OPEN)
        return pform

    def _print_Sequence(self, seq):
        from sympy.printing.pretty.stringpict import prettyForm, stringPict

        pforms = []
        if seq.start_trunc:
            pforms.append('...')

        for item, n1 in zip(seq.vals, seq.n):
            if pforms:
                pforms.append(', ')
            if n1 == 0:
                pforms.append('_')
            pform = self._print(item)
            pforms.append(pform)

        if seq.end_trunc:
            if pforms:
                pforms.append(', ')
            pforms.append('...')

        if not pforms:
            s = stringPict('')
        else:
            s = prettyForm(*stringPict.next(*pforms))

        s = prettyForm(*s.parens('{', '}', ifascii_nougly=True))
        return s

    def _print_Symbol(self, expr):

        expr = sym.Symbol(canonical_name(expr.name))
        parts = expr.name.split('_')
        if len(parts) >= 2:
            # Due to unicode limitations, Sympy cannot print a symbol
            # name with a double subscript.  As a work-around combine
            # the subscripts.  Note, Sympy converts 'v_C1' into
            # 'v_C_1' so we need to clean up.
            expr.name = parts[0] + '_' + ''.join(parts[1:])
        s = super(LcapyPrettyPrinter, self)._print_Symbol(expr)
        return s

    def _print_Piecewise(self, expr):

        from sympy.printing.pretty.stringpict import prettyForm

        if len(expr.args) > 1:
            return super(LcapyPrettyPrinter, self)._print_Piecewise(expr)

        ec = expr.args[0]

        pform = self._print(ec.expr)
        pform = prettyForm(*pform.right('  for '))
        pform = prettyForm(*pform.right(self._print(ec.cond)))
        return pform

    def _print_Heaviside(self, expr):

        from sympy.printing.pretty.stringpict import prettyForm

        if self._use_unicode:
            pform = self._print(expr.args[0])
            pform = prettyForm(*pform.parens(left='(', right=')'))
            pform = prettyForm(*pform.left(pretty_expr_map[sym.Heaviside]))
            return pform
        else:
            return self._print_Function(expr)

    def _print_UnitImpulse(self, expr):

        from sympy.printing.pretty.stringpict import prettyForm
        from sympy.printing.pretty.pretty_symbology import greek_unicode

        if self._use_unicode:
            pform = self._print(expr.args[0])
            pform = prettyForm(*pform.parens(left='[', right=']'))
            pform = prettyForm(*pform.left(greek_unicode['delta']))
            return pform
        else:
            return self._print_Function(expr)

    def _print_UnitStep(self, expr):

        from sympy.printing.pretty.stringpict import prettyForm

        if self._use_unicode:
            pform = self._print(expr.args[0])
            pform = prettyForm(*pform.parens(left='[', right=']'))
            pform = prettyForm(*pform.left('u'))
            return pform
        else:
            return self._print_Function(expr)

    def _print_dtrect(self, expr):

        from sympy.printing.pretty.stringpict import prettyForm

        if self._use_unicode:
            pform = self._print(expr.args[0])
            pform = prettyForm(*pform.parens(left='[', right=']'))
            pform = prettyForm(*pform.left('rect'))
            return pform
        else:
            return self._print_Function(expr)

    def _print_dtsign(self, expr):

        from sympy.printing.pretty.stringpict import prettyForm

        if self._use_unicode:
            pform = self._print(expr.args[0])
            pform = prettyForm(*pform.parens(left='[', right=']'))
            pform = prettyForm(*pform.left('sign'))
            return pform
        else:
            return self._print_Function(expr)


# print_function is a decorator to replace kwargs with the printer
# settings in __signature__

@print_function(LcapyStrPrinter)
def print_str(expr, **settings):
    """Convert expression into a string."""

    printer = LcapyStrPrinter(settings)
    return printer.doprint(expr)


@print_function(LcapyPrettyPrinter)
def pretty(expr, **settings):
    """Pretty print an expression."""

    printer = LcapyPrettyPrinter(settings)
    return printer.doprint(expr)


def pprint(expr, **kwargs):
    """Pretty print an expression.

    If have non-interactive shell a latex string is returned."""

    import sys

    # If interactive use pretty, otherwise use latex
    if hasattr(sys, 'ps1'):
        print(pretty(expr, **kwargs))
    else:
        print(latex(expr, **kwargs))


@print_function(LcapyLatexPrinter)
def latex(expr, **settings):

    printer = LcapyLatexPrinter(settings)
    string = printer.doprint(expr)

    match = func_pattern.match(string)
    if match is not None:
        # v_1(t) -> \operatorname{v\_1}\left( t \right)
        # operatorname requires amsmath so switch to mathrm
        string = r'\mathrm{%s}' % match.groups()[0].replace('\\_', '_')

    return string


from sympy import init_printing  # nopep8

init_printing(latex_printer=latex, pretty_printer=pretty,
              str_printer=print_str)

# See sympy/interactive/printing.py and IPython/core/formatters.py
# types.



"""This module handles the nodes for a circuit.

Copyright 2023 Michael Hayes, UCECE
"""

from .attrdict import AttrDict
from .schemmisc import Pos
from .node import Node


class Nodes(AttrDict):

    def add(self, node_name, cpt, cct):

        if node_name in self:
            node = self[node_name]
        else:
            node = Node(cct, node_name)
            self[node_name] = node
        node.append(cpt)
        return node

    def _remove(self, node_name):
        """Remove node from dict of nodes.  Note, it is best to use Node.remove;
        this will not remove the Node until all uses are removed."""

        node = self[node_name]
        if len(node.connected) != 0:
            raise ValueError('Cannot remove node ' + str(node_name) +
                             '; it is needed for ' +
                             ', '.join([str(name) for name in node.connected]))

        self.pop(node_name)

    def debug(self):

        s = ''
        for node in self.values():
            s += node.debug()
        return s

    def by_position(self, position):

        x, y = position

        for node in self.values():
            if abs(node.x - x) < 1e-5 and abs(node.y - y) < 1e-5:
                return node
        return None


def parse_nodes(nodesstr):

    from .utils import split_parens

    node_positions = {}

    # Ignore {}
    nodesstr = nodesstr[1:-1]
    entries = split_parens(nodesstr, ',')
    for entry in entries:
        if entry == '':
            continue
        parts = entry.split('@')
        node_name = parts[0].strip()
        # Ignore ()
        values = parts[1][1:-1]
        parts = values.split(',')
        x = float(parts[0])
        y = float(parts[1])
        pos = Pos(x, y)

        node_positions[node_name] = pos

    return node_positions



from .quantity import Quantity


class VoltageMixin(Quantity):

    quantity = 'voltage'
    quantity_label = 'Voltage'
    quantity_units = 'V'
    is_voltage = True
    is_signal = True
    is_undefined = False

    def cpt(self):
        from .oneport import V, Vac, Vdc

        if self.is_dc:
            return Vdc(self.expr)
        elif self.is_phasor_domain:
            return Vac(self, omega=self.omega)

        return V(self.expr)



"""
This module collects information about a circuit.

Copyright 2022--203 Michael Hayes, UCECE

"""


class Analysis:

    def __init__(self, cct, components):

        self.has_ic = False
        self.zeroic = True
        self.has_s = False
        self.has_transient = False
        self.ac_count = 0
        self.dc_count = 0
        self.causal = True
        self.reactive = False
        self.independent_sources = []
        self.dependent_sources = []
        self.control_sources = []
        self.reactances = []
        self.ics = []
        self.mutual_couplings = []

        for eltname, elt in cct.elements.items():
            if elt.need_control_current:
                self.control_sources.append(elt.args[0])
            if elt.has_ic is not None:
                if elt.has_ic:
                    self.has_ic = True
                    self.ics.append(eltname)
                if not elt.zeroic:
                    self.zeroic = False
            if elt.is_independent_source:
                self.independent_sources.append(eltname)
                if elt.has_s_transient:
                    self.has_s = True
                if elt.has_transient:
                    self.has_transient = True
                if elt.is_ac:
                    self.ac_count += 1
                if elt.is_dc:
                    self.dc_count += 1
                if not elt.is_causal:
                    self.causal = False
            if elt.is_dependent_source:
                self.dependent_sources.append(eltname)
            if elt.is_reactive:
                self.reactive = True
                self.reactances.append(eltname)
            if elt.is_mutual_coupling:
                self.mutual_couplings.append(eltname)

        num_sources = len(self.independent_sources)

        self.has_dc = self.dc_count > 0
        self.has_ac = self.ac_count > 0

        self.ac = self.ac_count > 0 and \
            (num_sources == self.ac_count) and not self.has_ic
        self.dc = self.dc_count > 0 and \
            (num_sources == self.dc_count) and not self.has_ic
        self.causal = self.causal and self.zeroic
        self.time_domain = not self.reactive and not self.has_s

        self.ivp = self.has_ic
        self.switching = components.switches != []

        if not self.reactive and self.has_ic:
            raise ValueError('Non-reactive component with initial conditions')

    def __repr__(self):

        return str(vars(self))



"""
This module collects collates the circuit components.

Copyright 2022 Michael Hayes, UCECE

"""


class Components:

    def __init__(self, cct):

        self.capacitors = []
        self.current_sources = []
        self.inductors = []
        self.resistors = []
        self.switches = []
        self.open_circuits = []
        self.ports = []
        self.transformers = []
        self.voltage_sources = []
        self.wires = []

        for eltname, elt in cct.elements.items():

            if elt.is_capacitor:
                self.capacitors.append(eltname)
            elif elt.is_current_source:
                self.current_sources.append(eltname)
            elif elt.is_inductor:
                self.inductors.append(eltname)
            elif elt.is_open_circuit:
                self.open_circuits.append(eltname)
            elif elt.is_port:
                self.ports.append(eltname)
            elif elt.is_resistor:
                self.resistors.append(eltname)
            elif elt.is_switch:
                self.switches.append(eltname)
            elif elt.is_transformer:
                self.transformers.append(eltname)
            elif elt.is_voltage_source:
                self.voltage_sources.append(eltname)
            elif elt.is_wire:
                self.wires.append(eltname)

    def __repr__(self):

        return str(vars(self))



from .quantity import Quantity


class VoltageSquaredMixin(Quantity):

    quantity = 'voltagesquared'
    quantity_label = 'Voltage^2'
    quantity_units = 'V^2'
    is_voltagesquared = True
    is_squared = True
    is_undefined = False



import numpy as np
from matplotlib.pyplot import subplots, style, savefig, show

from lcapy import *

h = exp(-t) * u(t)
h1 = h(n, method='impulse-invariance').subs(dt, 0.1)
# Same as impulse-invariance since there are no zeros
#h2 = h(n, method='matched-Z').subs(dt, 0.1)
h3 = h(n, method='bilinear').subs(dt, 0.1)
h4 = h(n, method='forward-euler').subs(dt, 0.1)
h5 = h(n, method='backward-euler').subs(dt, 0.1)

fig, axes = subplots(1, figsize=(6, 3.5))

h1.plot((-5, 15), axes=axes, label='impulse-invariance', color='C0')
#h2.plot((-5, 15), axes=axes, label='matched-Z', color='C1')
h3.plot((-5, 15), axes=axes, label='bilinear', color='C2')
h4.plot((-5, 15), axes=axes, label='forward-Euler', color='C3')
h5.plot((-5, 15), axes=axes, label='backward-Euler', color='C4')
axes.set_ylim(0, 1.5)

axes.legend()

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



"""This modules provides the DiscreteExpression class to provide
common methods for the discrete-time representations.

Copyright 2020 Michael Hayes, UCECE

"""

from .expr import Expr


class DiscreteExpression(Expr):

    """Superclass of discrete-time, discrete-frequency, and z-domain
    expressions."""

    pass



from matplotlib.pyplot import savefig
from lcapy import *

H = 10 * (s + 1) * (s + 2) / ((s - 3) * (s - 4))

H.nyquist_plot((-100, 100))

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')




"""This module provides differential equation support.

Copyright 2021 Michael Hayes, UCECE

"""

from .expr import equation, expr
from .texpr import TimeDomainExpression, texpr
import sympy as sym


class DifferentialEquation(TimeDomainExpression):

    def __init__(self, lhs, rhs, inputsym='x', outputsym='y', **assumptions):

        lhs = texpr(lhs, **assumptions)
        rhs = texpr(rhs, **assumptions)

        lhssymbols = lhs.symbols
        rhssymbols = rhs.symbols

        if inputsym not in rhssymbols:
            raise ValueError('Input symbol %s not in rhs %s' % (inputsym, rhs))
        if outputsym not in lhssymbols:
            raise ValueError('Output symbol %s not in lhs %s' %
                             (outputsym, lhs))

        cls = lhs.__class__
        eqn = sym.Eq(lhs.expr, rhs.expr, evaluate=False)
        super(DifferentialEquation, self).__init__(cls(eqn), **assumptions)
        self.inputsym = inputsym
        self.outputsym = outputsym

    def __eq__(self, x):

        if not isinstance(x, self.__class__):
            return False

        return self.lhs == x.lhs and self.rhs == x.rhs

    def transfer_function(self):
        """Create continuous-time transfer function assuming zero initial
        conditions."""
        from .sexpr import sexpr

        X = self.inputsym.upper()
        Y = self.outputsym.upper()

        Ys = self.LT(zero_initial_conditions=True).solve(Y)[0]
        Xs = sexpr(X + '(s)')

        result = Ys / Xs
        result.is_causal = True

        return result

    @property
    def lhs(self):
        """Return left-hand-side of differential equation."""

        return TimeDomainExpression(self.expr.lhs, **self.assumptions)

    @property
    def rhs(self):
        """Return right-hand-side of differential equation."""

        return TimeDomainExpression(self.expr.rhs, **self.assumptions)

    def lti_filter(self):
        """Create continuous-time linear time-invariant filter."""

        return self.transfer_function().lti_filter()

    def angular_frequency_response(self):
        """Return angular frequency response."""

        return self.transfer_function().angular_frequency_response()

    def frequency_response(self):
        """Return frequency response."""

        return self.transfer_function().frequency_response()

    def impulse_response(self):
        """Return impulse response."""

        return self.transfer_function().impulse_response()

    def step_response(self):
        """Return step response."""

        return self.transfer_function().step_response()

    def separate(self):
        """Rewrite differential equation so that input symbols are on the right
        and output symbols are on the left."""

        newlhs = 0
        newrhs = 0

        for term in self.lhs.as_ordered_terms():
            symbols = texpr(term).symbols
            if self.outputsym in symbols:
                newlhs += term
            if self.inputsym in symbols:
                newrhs -= term

        for term in self.rhs.as_ordered_terms():
            symbols = texpr(term).symbols
            if self.outputsym in symbols:
                newlhs -= term
            if self.inputsym in symbols:
                newrhs += term

        return self.__class__(newlhs, newrhs, **self.assumptions)

    def discretize(self, method='bilinear', alpha=0.5):
        """Convert to a difference equation.

        The default method is 'bilinear'.  Other methods are:
        'impulse-invariance' 'bilinear', 'tustin', 'trapezoidal'
        'generalized-bilinear', 'gbf' controlled by the parameter
        `alpha`, 'euler', 'forward-diff', 'forward-euler'
        'backward-diff', 'backward-euler' 'simpson', 'matched-Z',
        'zero-pole-matching'"""

        H = self.transfer_function()

        return H.discretize(method, alpha).difference_equation()



from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
from .admittancemixin import AdmittanceMixin
from .impedancemixin import ImpedanceMixin
from .transfermixin import TransferMixin
from .voltagesquaredmixin import VoltageSquaredMixin
from .currentsquaredmixin import CurrentSquaredMixin
from .admittancesquaredmixin import AdmittanceSquaredMixin
from .impedancesquaredmixin import ImpedanceSquaredMixin
from .powermixin import PowerMixin


quantities = {'voltage': VoltageMixin,
              'current': CurrentMixin,
              'admittance': AdmittanceMixin,
              'impedance': ImpedanceMixin,
              'transfer': TransferMixin,
              'voltagesquared': VoltageSquaredMixin,
              'currentsquared': CurrentSquaredMixin,
              'admittancesquared': AdmittanceSquaredMixin,
              'impedancesquared': ImpedanceSquaredMixin,
              'power': PowerMixin}



"""This module provides current support.

Copyright 2020--2023 Michael Hayes, UCECE

"""
from .deprecation import LcapyDeprecationWarning
from .expr import expr
from .state import state
from .classmap import domain_kind_to_symbol, domain_kind_quantity_to_class
from warnings import warn


def Iname(name, kind, cache=False):

    # Not caching is a hack to avoid conflicts of Vn1 with Vn1(s) etc.
    # when using subnetlists.  The alternative is a proper context
    # switch.  This would require every method to set the context.

    if kind in ('t', 'time'):
        name = name.lower()

    undef = domain_kind_to_symbol(kind, name)
    cls = domain_kind_quantity_to_class(kind, 'current')
    return cls(undef, cache=cache)


def Itype(kind):

    return domain_kind_quantity_to_class(kind, 'current')


def current(arg, **assumptions):

    expr1 = expr(arg, **assumptions)

    if 'nid' in assumptions:
        from .noisefexpr import FourierNoiseDomainCurrent
        from .noiseomegaexpr import AngularFourierNoiseDomainCurrent

        if expr1.is_fourier_domain or expr1.is_constant_domain:
            expr1 = FourierNoiseDomainCurrent(expr1)
        elif expr1.is_angular_fourier_domain:
            expr1 = AngularFourierNoiseDomainCurrent(expr1)
        else:
            raise ValueError(
                'Cannot represent noise current in %s domain' % expr1.domain)

    try:
        expr1 = expr1.as_current()
    except:
        raise ValueError('Cannot represent %s(%s) as current' %
                         (expr1.__class__.__name__, expr1))

    return expr1


def noisecurrent(arg, **assumptions):
    """Create a new noise current with specified amplitude spectral density."""

    nid = assumptions.get('nid', None)
    positive = assumptions.get('positive', True)

    return current(arg, nid=None, positive=positive, **assumptions)


def phasorcurrent(arg, omega=None, **assumptions):
    from .phasor import phasor

    return phasor(arg, omega, **assumptions).as_current()


def current_sign(I, is_source):
    """Manipulate the sign of the current according to the current sign
    convention specified by `state.current_sign_convention`."""

    sign_convention = state.current_sign_convention

    if sign_convention is None:
        sign_convention = 'hybrid'

        if is_source:
            state.sign_convention = sign_convention
            warn(
                """The default hybrid sign convention for currents is deprecated and will default to the passive sign convenention in the next version of Lcapy.  This only affects the sign of the current through sources.  For example, given the netlist

I1 1 0
R1 1 0

the hybrid sign convention gives `I1.i` as `I1` and the passive sign
convention gives `I1.i` as -I1.

To select the passive sign convention use:
`from lcapy import state; state.current_sign_convention = 'passive'`

To select the hybrid sign convention use:
`from lcapy import state; state.current_sign_convention = 'hybrid'`

""")

    if ((sign_convention == 'hybrid' and is_source)
            or sign_convention == 'active'):
        return -I
    return I



from lcapy import *
from lcapy.jomegaexpr import AngularFrequencyResponseDomainExpression
import unittest


class LcapyTester(unittest.TestCase):
    """Unit tests for lcapy"""

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            pprint(ans1)
            pprint(ans2)
            raise AssertionError(e)

    def test_laplace_jomega_transform(self):

        H = 1 / s
        P = H(jw)

        self.assertEqual(type(P), AngularFrequencyResponseDomainExpression, 'H(jw)')



from .immittancemixin import ImmittanceMixin
from .quantity import Quantity


class AdmittanceMixin(Quantity, ImmittanceMixin):

    quantity = 'admittance'
    quantity_label = 'Admittance'
    quantity_units = 'S'
    is_admittance = True
    is_immittance = True
    is_ratio = True
    is_undefined = False

    # Immittances derived from a realisable circuit will be causal but
    # non-causal immittances can also be constructed.  So this might
    # disappear.  An example non-causal impulse response is z(t) = R delta(t + T)
    # with an impedance Z(s) = R exp(s * T).  This has
    # a real part R * exp(re(s) * T) * cos(T * im(s))
    # and imaginary part R * exp(re(s) * T) * sin(T * im(s))

    is_always_causal = True

    @property
    def Y(self):
        """Admittance."""
        return self

    @property
    def Z(self):
        """Impedance."""

        from .impedance import impedance

        return impedance(1 / self)

    def __rtruediv__(self, x):
        """Reverse true divide"""

        # FIXME for things like 1 / (R * C).

        x = expr(x)
        if x.is_constant:
            from .impedance import impedance
            return impedance(x.expr / self.expr)
        return super(AdmittanceMixin, self).__rtruediv__(x)

    def cpt(self):
        from .oneport import G, C, L, Y
        from .symbols import s

        if self.is_number or self.is_dc:
            return G(self.expr)

        y = self.laplace() * s

        if y.is_number:
            return L((1 / y).expr)

        y = self.laplace() / s

        if y.is_number:
            return C(y.expr)

        return Y(self)


from .expr import expr  # nopep8



"""
Deprecation warning.

This is adapted from sympy/utilities/exceptions.py
"""

import warnings

from sympy.utilities.misc import filldedent


class LcapyDeprecationWarning(DeprecationWarning):
    r"""A warning for deprecated features of Lcapy.
    """

    def __init__(self, feature=None, last_supported_version=None,
                 useinstead=None, issue=None, deprecated_since_version=None):

        self.fullMessage = ""

        if not feature:
            raise ValueError(
                "feature is required argument of LcapyDeprecationWarning")

        if not deprecated_since_version:
            raise ValueError(
                "deprecated_since_version is a required argument of LcapyDeprecationWarning")

        self.fullMessage = "%s has been deprecated since Lcapy %s. " % \
            (feature, deprecated_since_version)

        if last_supported_version:
            self.fullMessage += ("It will be last supported in Lcapy "
                                 "version %s. ") % last_supported_version

        if useinstead:
            self.fullMessage += "Use %s instead. " % useinstead

        if issue is not None:
            self.fullMessage += ("See "
                                 "https://github.com/mph-/lcapy/issues/%d for more "
                                 "info. ") % issue

    def __str__(self):
        return '\n%s\n' % filldedent(self.fullMessage)

    def warn(self, stacklevel=2):
        # the next line is what the user would see after the error is printed
        # if stacklevel was set to 1. If you are writing a wrapper around this,
        # increase the stacklevel accordingly.
        warnings.warn(self, stacklevel=stacklevel)


# Python by default hides DeprecationWarnings, which we do not want.
warnings.simplefilter("once", LcapyDeprecationWarning)



"""This module provides discrete-time filter support.

Copyright 2021--2023 Michael Hayes, UCECE

"""

from .expr import expr, equation, ExprTuple, ExprList
from .functions import Function, Derivative, sin, exp
from .nexpr import DiscreteTimeDomainExpression
from .differenceequation import DifferenceEquation
from .discretetime import n, z, seq
from .sequence import Sequence
from .symbols import f
from .transfer import transfer
from .utils import isiterable
from .sym import oo, j, pi, dt
import sympy as sym
from warnings import warn


# TODO: convert lowpass form to highpass with H(-z)


class DLTIFilter(object):

    def __init__(self, b, a):
        """Create discrete-time filter where `b` is a list or array of
        numerator coefficients and `a` is a list or array of
        denominator coefficients.
        The order in a and b is according to z^0, z^(-1), ...

        Note, a recursive (IIR) filter with more than 3 coefficients
        for `a` is sensitive to truncation of the cofficients and can
        easily become unstable.  It is better to use a sequence
        of biquadratic filters.

        """

        if not isiterable(b):
            b = (b, )
        if not isiterable(a):
            a = (a, )

        self.a = ExprTuple(a)
        self.b = ExprTuple(b)

    def __repr__(self):
        """This is called by repr(expr).  It is used, e.g., when printing
        in the debugger."""

        return '%s(%s, %s)' % (self.__class__.__name__, self.b, self.a)

    @classmethod
    def from_ZPK(cls, Z, P, K=1):
        """Create LTIFilter given transfer function in zero-pole-gain form
        where `Z` is a list of zeroes, `P` is a list of poles, and `K`
        is a constant."""

        if not isiterable(Z):
            Z = (Z, )
        if not isiterable(P):
            P = (P, )

        N = expr(K)
        for z1 in Z:
            N *= (z - z1)
        b = N.coeffs()

        D = expr(1)
        for p in P:
            D *= (z - p)
        a = D.coeffs()

        return cls(b, a)

    @classmethod
    def from_transfer_function(cls, H, normalize=True):
        """Create LTIFilter given a transfer function.

        If `normalize` is True then normalize all the
        numerator and denominator coefficients by the first non-zero
        denominator coefficient.
        """

        H = transfer(H)
        if not H.is_rational_function:
            raise ValueError("Transfer function is not a rational function")

        N, D = H.as_N_D()

        nn = N.coeffs()
        dn = D.coeffs()

        if len(nn) > len(dn):
            # Have improper transfer function
            warn("System not causal")
            an = ExprList((len(nn) - len(dn)) * [0] + dn)
            bn = nn
        else:
            bn = ExprList((len(dn) - len(nn)) * [0] + nn)
            an = dn

        # Remove trailing zero coefficients.  This is equivalent
        # to calling cancel on the transfer function.
        while an[-1] == 0 and bn[-1] == 0:
            an = an[0:-1]
            bn = bn[0:-1]

        if normalize:

            # Use first non-zero coeff as normalizer to handle cases
            # where a[0] = 0.
            for norm in an:
                if norm != 0:
                    break
            bn = [bx / norm for bx in bn]
            an = [ax / norm for ax in an]

        return cls(bn, an)

    def transfer_function(self):
        """Return discrete-time transfer function:

                                    -1              -M
                Y(z)    b[0] + b[1]z  + ... + b[M] z
         H(z) = ---- = --------------------------------
                X(z)                -1              -N
                        a[0] + a[1]z  + ... + a[N] z

        where `M` is the degree of the numerator and `N` is the degree of
        the denominator.
        """

        from .sym import zsym

        a = self.a.sympy
        b = self.b.sympy

        N = len(a)
        M = len(b)

        # Numerator of H(z)
        numer = sym.S.Zero
        for i in range(M):
            numer += b[i] * zsym**(-i)

        # Denominator for H(z)
        denom = sym.S.Zero
        for k in range(N):
            denom += a[k] * zsym**(-k)

        Hz = transfer(numer, denom)
        Hz.is_causal = True
        return Hz

    def impulse_response(self):
        """Return discrete-time impulse response (transfer function) in time domain."""

        H = self.transfer_function()
        return H(n)

    def step_response(self):
        """Return discrete-time step response in time domain."""

        H = self.transfer_function()
        G = (H * z) / (z - 1)
        G.is_causal = H.is_causal

        return G(n)

    @property
    def is_moving_average(self):

        if len(self.a) != 1 or len(self.b) < 2:
            return False

        for m in range(len(self.b) - 1):
            if self.b[m + 1] != self.b[0]:
                return False
        return True

    def frequency_response(self, var=None, images=oo, **assumptions):
        """Return frequency response."""

        if self.is_moving_average:
            N = len(self.b)
            return sin(N * pi * f * dt) / sin(pi * f * dt) \
                * exp(-j * pi * f * (N - 1) * dt) * N * self.b[0] / self.a[0]

        return self.transfer_function().DTFT(var, images, **assumptions)

    def phase_response(self):

        if self.is_moving_average:
            N = len(self.b)
            return -pi * f * (N - 1) * dt

        H = self.frequency_response()
        return H.phase

    def group_delay(self):

        phi = self.phase_response()
        tau = -Derivative(phi, f) / (2 * pi)
        return tau.simplify().general()

    def phase_delay(self):

        phi = self.phase_response()
        tau = -phi / (2 * pi * f)
        return tau.simplify().general()

    def difference_equation(self, inputsym='x', outputsym='y'):
        """Return difference equation in the form:

         a[0]*y[n] = b[0]*x[n] + b[1]*x[n-1] + ... + b[M]*x[n-M]
                               - a[1]*y[n-1] - ... - a[N]*y[n-N]

        where `M` is the degree of the transfer function numerator and
        `N` is the degree of the transfer function denominator.

        """

        lhs = 0 * n
        rhs = 0 * n

        for m, bn in enumerate(self.b):
            rhs += bn * Function(inputsym)(n - m)

        use_lhs = True
        for m, an in enumerate(self.a):
            x = an * Function(outputsym)(n - m)
            if use_lhs:
                lhs += x
            else:
                rhs -= x
            if an != 0:
                use_lhs = False

        return DifferenceEquation(lhs, rhs, inputsym, outputsym)

    def zdomain_initial_response(self, ic=None, xic=None, left=True,
                                 inputsym='x', outputsym='y'):
        """
           Return z-domain response due to initial conditions.
           zdomain_initial_response(ic=None, xic=None, left=True, inputsym='x', outputsym='y')
           left=True (default)
              ic : list with initial values y[-1], y[-2], ...,
              xic : list with initial values x[-1], x[-2], ...,
           left=False
              ic : list with initial values y[0], y[1], ...
              xic : list with initial values x[0], x[1], ...
           If no ic or xic are given symbolic expressions are used
        """

        from .sym import zsym

        Nl = len(self.a)

        # default values for ic
        if ic is None:
            ic = []
            n_pos = range(-1, -Nl, -1)
            if not left:
                n_pos = range(0, Nl - 1)
            for k in n_pos:
                ic += [sym.Symbol('%s[%i]' % (outputsym, k))]
            ic = tuple(ic)

        # default values for xic
        if xic is None:
            xic = []
            n_pos = range(-1, -Nl, -1)
            if not left:
                n_pos = range(0, Nl - 1)
            for k in n_pos:
                xic += [sym.Symbol('%s[%i]' % (inputsym, k))]
            xic = tuple(xic)

        if not isiterable(ic):
            ic = (ic, )
        if not isiterable(xic):
            xic = (xic, )

        Nl = len(self.a)
        if len(ic) != Nl - 1:
            raise ValueError('Expecting %d initial conditions for ic, got %d'
                             % (Nl - 1, len(ic)))
        if len(xic) != Nl - 1:
            raise ValueError('Expecting %d initial conditions for xic, got %d'
                             % (Nl - 1, len(ic)))

        # Initial condition y[-1], y[-2], .....
        if left:
            # Denominator for Yi(z)
            denom = self.a[0] * z**0
            num = 0 * z
            for k in range(1, Nl):
                az = self.a[k] * z**(-k)
                denom += az
                # Numerator for Yi(z)
                y0 = 0 * z
                x0 = 0 * z
                for i in range(0, k):
                    y0 += ic[i] * z**(i + 1)
                    x0 += xic[i] * z**(i + 1)
                # ic conditions
                num -= az * y0
                # xic conditions
                if len(self.b) >= k + 1:
                    num += x0 * self.b[k] * z**(-k)

            # Collect with respect to positive powers of the variable z
            num = num.sympy * zsym**(Nl - 1)
            denom = denom.sympy * zsym**(Nl - 1)

        # Initial condition y[0], y[1], .....
        else:
            # Denominator for Yi(z)
            denom = self.a[Nl - 1]
            num = 0 * z
            for k in range(0, Nl - 1):
                az = self.a[k] * z**(Nl - 1 - k)
                denom += az
                # Numerator for Yi(z)
                y0 = 0 * z
                x0 = 0 * z
                for i in range(0, Nl - 1 - k):
                    y0 += ic[i] * z ** (-i)
                    x0 -= xic[i] * z ** (-i)
                # ic conditions
                num += az * y0
                # xic conditions
                if len(self.b) >= k + 1:
                    num += x0 * self.b[k] * z**(Nl - 1 - k)

        num = sym.collect(sym.expand(num), zsym)
        denom = sym.collect(sym.expand(denom), zsym)

        Yzi = expr(num / denom)
        Yzi.is_causal = True

        return Yzi

    def initial_response(self, ic=None, xic=None, left=True, inputsym='x', outputsym='y'):
        """
           initial_response(ic=None, xic=None, left=True, inputsym='x', outputsym='y')
           Return response due to initial conditions in the time domain.
           left=True (default)
              ic : list with initial values y[-1], y[-2], ...
              xic : list with initial values x[-1], x[-2], ...
           left=False
              ic : list with initial values y[0], y[1], ...
              xic : list with initial values x[0], x[1], ...
           If no ic or xic are given symbolic expressions are used
        """

        Yzi = self.zdomain_initial_response(
            ic=ic, xic=xic, left=left, inputsym=inputsym, outputsym=outputsym)
        return Yzi(n)

    def response(self, x, ic=None, ni=None):
        """
        response(x, ic=None, ni=None)
        Calculate response of filter to input `x` at time indexes specified by `ni`,
        given a list of initial conditions `ic`.

        - `x` can be an expression, a sequence, or a list/array of values.
        - If `ni` is a tuple, this specifies the first and last (inclusive) time index.
          Otherwise, it should be list of integers, monotonically increasing by 1.
          If `ni` is `None`, or unspecified, it defaults to `(-len(ic), 10)`.
        - `ic` specifies the initial conditions, `[y[-1], y[-2], ...]`.
        - If `ic` is `None` the initial conditions are assumed to be zero.
        """

        from numpy import arange, ndarray

        if (ic is None and ni is None and isinstance(x, ndarray)
                and self.a.symbols == {} and self.b.symbols == {}):

            from scipy.signal import lfilter

            # Calculate numerical response
            b = self.b.fval
            a = self.a.fval
            return lfilter(b, a, x)

        if ic is None:
            ic = [0] * (len(self.a) - 1)

        if not isiterable(x) and not isinstance(x, DiscreteTimeDomainExpression):
            x = (x, )

        if not isiterable(ic):
            ic = (ic, )

        if isinstance(x, (tuple, list, ndarray)):
            x = seq(x)
        elif not isinstance(x, (Sequence, DiscreteTimeDomainExpression)):
            raise ValueError(
                'The input x must be a scalar, tuple, sequence, nexpr, list, or array')

        Ni = len(ic)

        if Ni != len(self.a) - 1:
            raise ValueError(
                "Expected %d initial conditions, got %d" % (len(self.a) - 1, Ni))

        if ni is None:
            ni = (-Ni, 10)

        if isinstance(ni, tuple):
            ni = arange(ni[0], ni[1] + 1)

        Nn = ni[-1] + 1

        # Order right hand side
        Nr = len(self.b)

        y_tot = list(ic[-1::-1]) + [0] * Nn

        a_r = self.a[-1:-1 - Ni:-1]
        for i in range(Nn):
            # Get previous y vals (sliding window)
            pre_y = y_tot[i:i + Ni]

            # Calculate rhs of new value
            if isinstance(x, Sequence):
                rhs = sum(self.b[l] * x[i - l] for l in range(Nr))
            else:
                rhs = sum(self.b[l] * x(i - l) for l in range(Nr))

            # Add lhs
            y_tot[i + Ni] = -1 / self.a[0] * \
                sum(csi * ysi for csi, ysi in zip(a_r, pre_y)) + \
                rhs / self.a[0]

        # Number of zeros to prepend
        Nz = 0
        if ni[0] < -Ni:
            Nz = -(Ni + ni[0])
            y_tot = Nz * [0] + y_tot

        ret_seq = seq(y_tot[ni[0] + Ni + Nz:], tuple(ni))

        return ret_seq

    def subs(self, *args, **kwargs):

        a = self.a.subs(*args, **kwargs)
        b = self.b.subs(*args, **kwargs)

        return self.__class__(b, a)

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self

    @property
    def is_stable(self):
        """Return True if all the poles of the filters transfer function
        are within the unit circle.

        See also is_marginally_stable."""

        return self.transfer_function().is_stable

    @property
    def is_marginally_stable(self):
        """Return True if all the poles of the filters transfer function
        are within or on the unit circle.

        See also is_stable."""

        return self.transfer_function().is_marginally_stable

    def inverse(self):
        """Create inverse filter.  Note, in practice, an inverse filter
        is sensitive to noise.  It is better to use a Wiener filter."""

        return self.__class__(self.a, self.b)



"""This module provides the PhasorDomain and PhasorRatioDomain classes
 to represent phasors and ratios of phasors for AC analysis.

A phasor represents the amplitude and phase for a single sinusoid.  By
default the angular frequency is omega_0 but it can be any number or
symbol.

Phasors straddle the time and frequency domains and hence the
confusing phasor classes.

A phasor is described by an amplitude and phase.  There is also an
implicit frequency.  The amplitude and phase are usually functions of
frequency.  A phasor is useful to describe an AC voltage or current
signal.

Lcapy considers V(jw) or I(jw) a generic phasor but V(3j) or I(3j) a
specific phasor.  Similarly, X(jw) is a generic phasor ratio and X(3j)
is a specific phasor ratio, where X denotes an immittance or transfer
function.

A ratio of two phasors (of the same frequency) is not a phasor.  The
ratio of two generic phasors is a frequency domain quantity.  The
frequency dependence is explicit.  A phasor ratio is useful to
describe an immittance or transfer function.  These are frequency
domain concepts.  A phasor ratio can be inferred from the Laplace
domain by substituting jomega (or jw) for s, where omega is the
angular frequency of the phasor.

Copyright 2014--2023 Michael Hayes, UCECE

"""

from __future__ import division
from .acdc import ACChecker
from .domains import PhasorDomain, PhasorRatioDomain
from .sym import j, omegasym, fsym, pi
from .expr import expr
from .functions import sin, cos, exp, sqrt
from .expr import Expr
from .omegaexpr import omega
from .units import u as uu
from sympy import Expr as symExpr
from warnings import warn


__all__ = ('PhasorDomainExpression',
           'PhasorRatioDomainExpression', 'phasor', 'phasor_ratio')

# The phasor domain is different from the Fourier and Laplace domain
# since there is an implicit angular frequency.  This is only needed
# for voltage and current (vorrent) expressions.

# The phasor domain immittance can be found from the Laplace domain
# impedance/admittance by substituting s with j * omega.  In many
# cases the result is the same as the Fourier domain immittance but
# not always.  The most important case is the impedance of a
# capacitor.  In the Laplace domain it is 1/(s C); in the phasor
# domain it is 1/(j omega C); in the Fourier domain it is 1/(2 j omega
# C) + delta(omega) / (2 C).


class PhasorExpression(Expr):

    @property
    def omega(self):
        """Return angular frequency."""

        return self.assumptions.get('omega', None)

    @property
    def var(self):
        """Return underlying variable as sympy expression."""

        if self.omega == omegasym:
            return omegasym

        if self.omega == 2 * pi * fsym:
            return fsym

        return None

    def phasor(self, **assumptions):
        """Convert to phasor representation."""

        ass = self.assumptions.copy()
        assumptions = ass.merge(**assumptions)

        return self.__class__(self, **assumptions)

    def rms(self):
        """Return root mean square."""

        return abs(self) * sqrt(2) / 2

    def _compatible_phasors(self, x):

        if self.domain != x.domain:
            return False

        if not hasattr(x, 'omega'):
            return True

        if self.omega == x.omega:
            return True

        raise ValueError('Incompatible phasor angular frequencies %s and %s' %
                         (self.omega, x.omega))

    def _add_compatible_domains(self, x):

        return self._compatible_phasors(x)

    def _div_compatible_domains(self, x):

        if (self.is_constant_domain or x.is_constant_domain):
            return True

        return self._compatible_phasors(x)

    def _mul_compatible_domains(self, x):

        if (self.is_constant_domain or x.is_constant_domain):
            return True

        if self._compatible_phasors(x):
            return True

        if (self.is_phasor_ratio_domain and x.is_phasor_domain and
                self.omega == x.omega):
            return True
        if (self.is_phasor_domain and x.is_phasor_ratio_domain and
                self.omega == x.omega):
            return True

        # Allow phasor(x) * omega, etc.
        if (self.is_phasor_domain and x.is_angular_fourier_domain and
                self.omega == x.var):
            return True
        if (self.is_phasor_domain and x.is_fourier_domain and
                self.omega == 2 * pi * x.var):
            return True

        if (self.is_phasor_ratio_domain and x.is_angular_fourier_domain and
                self.omega == x.var):
            return True
        if (self.is_phasor_ratio_domain and x.is_fourier_domain and
                self.omega == 2 * pi * x.var):
            return True

        return False

    def subs(self, *args, safe=False, **kwargs):
        """Substitute variables in expression, see sympy.subs for usage."""

        # Handle omega -> 2 * pi * f conversions to Fourier domain.
        if len(args) == 1 and args[0] == 2 * pi * fsym:
            from .fexpr import FourierDomainExpression

            if self.var != omegasym:
                raise ValueError('Expecting omega for self.var')

            # Could check for expressions that are known to fail,
            # such as 1 / (j * omega).
            if not safe:
                warn("""
Converting to Fourier domain via phasor domain may not give correct answer.   It is safer to convert to time domain then to Fourier domain.""")

            return FourierDomainExpression(self.sympy.subs(omegasym, args[0].sympy)).as_quantity(self.quantity)

        return super(PhasorExpression, self).subs(*args, **kwargs)


class PhasorDomainExpression(PhasorDomain, PhasorExpression):
    """This is a phasor domain base class for voltages and currents."""

    def __init__(self, val, **assumptions):

        if isinstance(val, PhasorExpression):
            assumptions['omega'] = val.omega
        elif 'omega' not in assumptions or assumptions['omega'] is None:
            assumptions['omega'] = omegasym

        if hasattr(assumptions['omega'], 'expr'):
            assumptions['omega'] = assumptions['omega'].expr

        assumptions['ac'] = True
        super(PhasorExpression, self).__init__(val, **assumptions)

    def _div_domain(self, x):

        if x.is_phasor_domain:
            return 'phasor ratio'

        return self.domain

    def _class_by_quantity(self, quantity, domain=None):

        if quantity == 'undefined':
            if domain == 'phasor ratio':
                return PhasorRatioDomainExpression
            else:
                return PhasorDomainExpression

        return super(PhasorDomainExpression, self)._class_by_quantity(quantity, domain)

    def as_expr(self):
        return PhasorDomainExpression(self)

    @classmethod
    def from_time(cls, expr, omega=None, **assumptions):

        from .symbols import t

        if expr.is_admittance or expr.is_impedance or expr.is_transfer:
            warn('Should convert %s expression to Laplace-domain first.' %
                 expr.quantity)

        assumptions['ac'] = True

        if expr.is_transform_domain:
            warn('Converting %s-domain to time-domain first.' % expr.domain)
            expr = expr.time()

        check = ACChecker(expr, t)
        if not check.is_ac:
            raise ValueError(
                'Do not know how to convert %s to phasor.  Expecting an AC signal.' % expr)

        if omega is not None and check.omega != omega:
            raise ValueError('Expecting omega=%s, found omega=%s.' %
                             (omega, check.omega))

        if check.omega == 0:
            warn('DC phasor.')

        result = check.amp * exp(j * check.phase)
        assumptions['omega'] = check.omega
        assumptions['complex_signal'] = check.is_complex

        return cls.change(expr, result, domain='phasor', **assumptions)

    @classmethod
    def from_constant(cls, expr, omega=None, **assumptions):

        return cls(expr, omega=omega)

    def time(self, **assumptions):
        """Convert to time domain representation."""
        from .symbols import t

        omega1 = self.omega
        if isinstance(omega1, Expr):
            # TODO: Fix inconsistency.  Sometimes omega is a symbol.
            omega1 = omega1.expr

        if self.is_complex_signal:
            result = self.expr * exp(j * omega1 * t)
        else:
            result = self.real.expr * \
                cos(omega1 * t) - self.imag.expr * sin(omega1 * t)

        return TimeDomainExpression(result).as_quantity(self.quantity)

    def laplace(self, **assumptions):
        """Convert to Laplace domain representation."""

        return self.time().laplace()

    def angular_fourier(self, **assumptions):
        """Angular Fourier transform."""

        if self.has(omega):
            warn('Expression contains omega, should substitute with a different symbol.')

        return self.time().angular_fourier()

    def fourier(self, **assumptions):
        """Fourier transform."""

        return self.time().fourier()

    def plot(self, **kwargs):
        """Plot polar diagram."""

        from .plot import plot_phasor

        return plot_phasor(self, **kwargs)


class PhasorRatioDomainExpression(PhasorRatioDomain, PhasorExpression):
    """This represents the ratio of two-phasors; for example
    an impedance, an admittance, or a transfer function."""

    is_phasor_ratio_domain = True
    is_transform_domain = True

    def __init__(self, val, **assumptions):

        if isinstance(val, PhasorExpression):
            assumptions['omega'] = val.omega
        elif 'omega' not in assumptions or assumptions['omega'] is None:
            assumptions['omega'] = omegasym

        if isinstance(val, Expr):
            ass = val.assumptions.copy()
            ass = ass.merge(**assumptions)
        else:
            ass = assumptions

        super(PhasorExpression, self).__init__(val, **ass)

    def _mul_domain(self, x):

        if x.is_phasor_domain:
            return 'phasor'

        return self.domain

    def _class_by_quantity(self, quantity, domain=None):

        if quantity == 'undefined':
            return PhasorRatioDomainExpression

        return super(PhasorRatioDomainExpression, self)._class_by_quantity(quantity, domain)

    @classmethod
    def from_laplace(cls, expr, **assumptions):

        from .sym import ssym

        ass = expr.assumptions.copy()
        ass = ass.merge(**assumptions)
        omega = ass.pop('omega', None)

        if omega is None:
            omega = omegasym
        elif hasattr(omega, 'expr'):
            omega = omega.expr

        if expr.is_voltage or expr.is_current:
            warn('Should convert %s expression to time-domain first.' %
                 expr.quantity)

        # Substitute jw for s
        result = expr.laplace(**ass)
        result2 = result.expr.replace(ssym, j * omega)

        quantity = expr.quantity
        if quantity == 'undefined':
            cls = PhasorRatioDomainExpression
        else:
            cls = expr._class_by_domain('phasor ratio')

        ret = cls(result2, omega=omega, **ass)
        return ret

    @property
    def domain_label(self):

        from .sym import fsym

        # TODO, remove this say by having PhasorRatioDomainExpression
        # and PhasorAngularFrequencyResponseDomainExpression classes.

        if self.var == fsym:
            return 'Frequency'
        else:
            return 'Angular frequency'

    @property
    def domain_units(self):

        from .sym import fsym
        from .units import u as uu

        # TODO, remove this say by having PhasorRatioDomainExpression
        # and PhasorAngularFrequencyResponseDomainExpression classes.

        if self.var == fsym:
            return uu.Hz
        else:
            return uu.rad / uu.s

    def time(self, **assumptions):
        """Convert to time domain representation."""

        return self.laplace().time()

    def laplace(self, **assumptions):
        """Convert to Laplace domain representation."""

        from .sym import ssym
        from .sexpr import LaplaceDomainExpression

        ass = self.assumptions.copy()
        assumptions = ass.merge(**assumptions)

        if self.var == omegasym:
            result = self.expr.replace(omegasym, ssym / j)
        elif self.var == fsym:
            result = self.expr.replace(fsym, ssym / (j * 2 * pi))
        else:
            raise ValueError(
                'Cannot convert PhasorRatioDomain to LaplaceDomain for variable %s' % self.var)

        return LaplaceDomainExpression(result, **assumptions).as_quantity(self.quantity)

    def angular_fourier(self, **assumptions):
        """Angular Fourier transform."""

        if self.var is omega:
            # TODO, add warning.
            pass

        return self.time().angular_fourier()

    def fourier(self, **assumptions):
        """Fourier transform."""

        return self.time().fourier()

    def as_expr(self):
        return PhasorRatioDomainExpression(self)

    def plot(self, fvector=None, **kwargs):
        """Plot frequency response.

        If the dependent variable is omega, fvector specifies the angular
        frequencies.  If it is a tuple, it sets the angular frequency
        limits.

        If the dependent variable is f, fvector specifies the linear
        frequencies.  If it is a tuple, it sets the linear frequency
        limits."""

        from .plot import plot_angular_frequency, plot_frequency
        from .sym import fsym

        if self.omega.is_constant():
            raise ValueError('Cannot plot at single frequency')

        if self.var == fsym:
            return plot_frequency(self, fvector, **kwargs)

        return plot_angular_frequency(self, fvector, **kwargs)

    def bode_plot(self, fvector=None, unwrap=True, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Bode
        plot (but without the straight line approximations).  fvector
        specifies the frequencies.  If it is a tuple (m1, m2), it sets the
        frequency limits as (10**m1, 10**m2).

        `unwrap` controls phase unwrapping (default True).
        """

        from .plot import plot_bode

        return plot_bode(self, fvector, unwrap=unwrap, **kwargs)


def phasor(arg, omega=None, **assumptions):
    """Create phasor.

    If arg has the form A * cos(w * t + phi) the phasor
    A * exp(j * phi) of angular frequency w is returned.

    If arg has the form A * sin(w * t + phi) the phasor
    A * exp(j * (phi - pi / 2)) of angular frequency w is returned.

    If arg is a constant C, the phasor C is created with omega as the
    angular frequency (this defaults to 'omega' if not specified).

    """

    arg = expr(arg)

    if arg.is_time_domain:
        # Expecting AC signal.
        return PhasorDomainExpression.from_time(arg, omega=omega, **assumptions)
    else:
        # Expecting phasor (complex amplitude)
        return PhasorDomainExpression(arg, omega=omega, **assumptions)


def phasor_ratio(arg, omega=None, **assumptions):
    """Create phasor ratio.

    If arg is a constant C, the phasor ratio C is created with omega as the
    angular frequency (this defaults to 'omega' if not specified).

    """

    arg = expr(arg)

    return PhasorRatioDomainExpression(arg, omega=omega, **assumptions)


from .texpr import TimeDomainExpression  # nopep8
from .expr import Expr  # nopep8



"""This module defines the StateSpace class for representing a linear
discrete time-invariant system as a state-space model.

Copyright 2021 Michael Hayes, UCECE

"""

from .cache import cached_property
from .matrix import Matrix
from .zmatrix import ZDomainMatrix
from .nmatrix import DiscreteTimeDomainMatrix
from .statespacebase import StateSpaceBase
from .nexpr import n, nexpr
from .equation import Equation
from .sym import zsym
import sympy as sym

# TODO: reachability implies controllability but controllability only
# implies reachability if A matrix is full rank.


class DTStateSpace(StateSpaceBase):
    """Discrete-time linear time-invariant state space model."""

    @property
    def u(self):
        """Input vector."""
        if self._u is None:
            self._u = DiscreteTimeDomainMatrix([nexpr('u_%d(n)' % i) for i in
                                                range(self.Nu)])
        return self._u

    @property
    def x(self):
        """State variable vector."""
        if self._x is None:
            self._x = DiscreteTimeDomainMatrix([nexpr('x_%d(n)' % i) for i in
                                                range(self.Nx)])
        return self._x

    @cached_property
    def xnext(self):
        """Time derivative of state variable vector."""
        return self.x.next_timestep()

    @property
    def x0(self):
        """State variable initial value vector."""
        if self._x0 is None:
            self._x0 = DiscreteTimeDomainMatrix([nexpr('x_0_%d(n)' % i) for i in
                                                 range(self.Nx)])
        return self._x0

    @property
    def y(self):
        """Output vector."""
        if self._y is None:
            self._y = DiscreteTimeDomainMatrix([nexpr('y_%d(n)' % i) for i in
                                                range(self.Ny)])
        return self._y

    @property
    def U(self):
        """Z transform of input vector."""
        return ZDomainMatrix(self.u.ZT())

    @property
    def X(self):
        """Z transform of state-variable vector."""
        return ZDomainMatrix(self.x.ZT())

    @property
    def Y(self):
        """Z transform of output vector."""
        return ZDomainMatrix(self.y.ZT())

    @cached_property
    def H(self):
        """X(z) / U(z)"""
        return ZDomainMatrix(self.Phi * self._B)

    @property
    def h(self):
        """ILT{X(z) / U(z)}"""
        return DiscreteTimeDomainMatrix(self.H.ILT(causal=True))

    @cached_property
    def G(self):
        """System transfer functions.
        For a SISO system, use G[0].
        """
        return ZDomainMatrix(self._C * self.H + self._D)

    def state_equations(self):
        """System of first-order differential state equations:

        x[n + 1] = A x[n] + B u[n]

        where x is the state vector and u is the input vector.
        """

        if len(self.u) == 0 and len(self.x) == 0:
            return Equation(None, None)

        if len(self.u) == 0:
            return Equation(self.xnext.sympy,
                            sym.MatMul(self._A.sympy, self.x.sympy))

        if len(self.x) == 0:
            return Equation(self.xnext.sympy,
                            sym.MatMul(self._B.sympy, self.u.sympy))

        return Equation(self.xnext.sympy,
                        sym.MatAdd(sym.MatMul(self._A.sympy, self.x.sympy),
                                   sym.MatMul(self._B.sympy, self.u.sympy)))

    def output_equations(self):
        """System of output equations:

        y[n] = C x[n] + D u[n]

        where y is the output vector, x is the state vector and u is
        the input vector.

        """
        if len(self.u) == 0:
            return Equation(self.y,
                            sym.MatMul(self._C.sympy, self.x.sympy))

        if len(self.x) == 0:
            return Equation(self.y,
                            sym.MatMul(self._D.sympy, self.u.sympy))

        return Equation(self.y,
                        sym.MatAdd(sym.MatMul(self._C.sympy, self.x.sympy),
                                   sym.MatMul(self._D.sympy, self.u.sympy)))

    @property
    def g(self):
        """System impulse responses."""
        return DiscreteTimeDomainMatrix(self.G.IZT(causal=True))

    @cached_property
    def Phi(self):
        """z-domain state transition matrix."""

        M = ZDomainMatrix(sym.eye(self.Nx) * zsym - self._A.sympy)
        return ZDomainMatrix(M.inv())

    @cached_property
    def phi(self):
        """State transition matrix."""
        return DiscreteTimeDomainMatrix(self.Phi.ILT(causal=True))

    def characteristic_polynomial(self):
        """Characteristic polynomial (aka system polynomial).

        `lambda(z) = |z * I - A|`"""

        M = ZDomainMatrix(sym.eye(self.Nx) * zsym - self._A.sympy)
        return ZDomainExpression(M.det()).simplify()

    @cached_property
    def P(self):
        """Characteristic polynomial (aka system polynomial).

        `lambda(z) = |z * I - A|`"""

        return self.characteristic_polynomial()

    @cached_property
    def Lambda(self):
        """Diagonal matrix of eigenvalues."""

        # Perhaps faster to use diagonalize
        # E, L = self.A.diagonalize()
        # return L

        e = self.eigenvalues
        return ZDomainMatrix(sym.diag(*e))

    @cached_property
    def M(self):
        """Modal matrix (eigenvectors of A)."""

        E, L = self._A.diagonalize()

        return ZDomainMatrix(E)

    @cached_property
    def controllability_gramian(self):
        """Controllability gramian matrix."""

        from scipy import linalg

        B = self.B.evaluate()
        Q = -B @ B.T

        # Find Wc given A @ Wc + Wc @ A.T = Q
        # Wc > o if (A, B) controllable
        Wc = linalg.solve_discrete_lyapunov(self.A.evaluate(), Q)

        # Wc should be symmetric positive semi-definite
        Wc = (Wc + Wc.T) / 2

        return Matrix(Wc)

    @property
    def Wc(self):
        """Controllability gramian matrix."""
        return self.controllability_gramian

    @property
    def reachability_gramian(self):
        """Reachability gramian matrix.  This is equivalent to the
        controllability gramian matrix for a linear time independent
        system provided A is not singular."""

        return self.controllability_gramian

    @property
    def Wr(self):
        """Reachability gramian matrix."""
        return self.reachability_gramian

    @cached_property
    def observability_gramian(self):
        """Observability gramian matrix."""

        from scipy import linalg

        C = self.C.evaluate()
        Q = -C.T @ C

        # Find Wo given A.T @ Wo + Wo @ A = Q
        # Wo > o if (C, A) observable
        Wo = linalg.solve_discrete_lyapunov(self.A.evaluate().T, Q)

        # Wo should be symmetric positive semi-definite
        Wo = (Wo + Wo.T) / 2

        return Matrix(Wo)

    @property
    def Wo(self):
        """Observability gramian matrix."""
        return self.observability_gramian

    def minimum_energy_input(self, steps, xfinal, xinitial=None):
        """Determine minimum energy (least norm) input vector that results in
        state `xfinal` in `steps` time steps from initial state
        `xinitial`."""

        if xinitial is None:
            xinitial = self.x0
        else:
            xinitial = Matrix(xinitial)

        xfinal = Matrix(xfinal)

        if steps < 0:
            raise ValueError('steps must be positive')

        C = self.controllability_matrix_steps(steps)

        xdiff = xfinal - xinitial

        uflip = C.T * (C * C.T).inv() * xdiff
        return uflip[::-1, :]

    def minimum_energy(self, steps, xfinal, xinitial=None):
        """Determine minimum energy that results in state `xfinal` in `steps`
        time steps from initial state `xinitial`."""

        if xinitial is None:
            xinitial = self.x0
        else:
            xinitial = Matrix(xinitial)

        xfinal = Matrix(xfinal)

        if steps < 0:
            raise ValueError('steps must be positive')

        C = self.controllability_matrix_steps(steps)

        xdiff = xfinal - xinitial

        return (xdiff.T * (C * C.T).inv() * xdiff)[0]

    def state_transfer(self, u, xinitial=None):
        """Return transitioned state given inputs `u` specified as a list of
        column vectors; one for each time step."""

        if xinitial is None:
            xinitial = self.x0
        else:
            xinitial = Matrix(xinitial)

        u = Matrix(u)
        steps = u.shape[0]

        C = self.controllability_matrix_steps(steps)

        xfinal = self.A**steps * xinitial + C * u[::-1, :]
        return xfinal


from .zexpr import ZDomainExpression  # nopep8



"""This module provides the TimeDomainExpression class to represent
time domain expressions.

Copyright 2014--2023 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import TimeDomain
from .expr import Expr, expr_make
from .functions import exp, integrate
from .state import state, validate
from .sym import fsym, omegasym, ssym, tsym, j, oo
from .laplace import laplace_transform
from .fourier import fourier_transform
from .units import u as uu
from sympy import Heaviside, limit, Expr as symExpr


__all__ = ('TimeDomainExpression', 'texpr', )


class TimeDomainExpression(TimeDomain, Expr):
    """t-domain expression or symbol."""

    var = tsym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)
        assumptions['real'] = True
        super(TimeDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr

        if check and not expr.has(tsym):
            if expr.has(fsym):
                validate(state.f_in_t,
                         't-domain expression %s depends on f' % expr)
            if expr.has(ssym):
                validate(state.s_in_t,
                         't-domain expression %s depends on s' % expr)
            if expr.has(omegasym):
                validate(state.w_in_t,
                         't-domain expression %s depends on omega' % expr)

    def _mul_compatible_domains(self, x):

        if self.domain == x.domain:
            return True

        return x.is_constant_domain

    def _mul_dubious(self, x):

        # Allow rect(t) * rect(t) etc.
        if not (self.is_signal or self.is_immittance or
                x.is_signal or x.is_immittance):
            return False

        # Allow v*v.
        return x.is_time_domain and not (self.is_signal and x.is_signal)

    def _div_compatible_domains(self, x):

        if self.domain == x.domain:
            return True

        return x.is_constant_domain

    @property
    def abs(self):
        """Return absolute value."""

        return self.__class__(abs(self.expr), **self.assumptions)

    def as_expr(self):
        return TimeDomainExpression(self)

    @property
    def energy(self):
        """Return signal energy."""

        # TODO: teach SymPy improper integral of rect and sinc functions

        # Only need abs for complex signal.
        E = integrate(abs(self)**2, (t, -oo, oo))

        # FIXME: result should not be a time-domain expression
        return E

    def infer_assumptions(self):

        self.assumptions.infer_from_expr(self)

    def LT(self, evaluate=True, zero_initial_conditions=None, **assumptions):
        """Determine one-sided Laplace transform with 0- as the lower limit.

        By default initial conditions are assumed to be zero.  This
        can be controlled by `zero_initial_conditions`.

        This is an alias for laplace.

        """
        if zero_initial_conditions is None:
            zero_initial_conditions = state.zero_initial_conditions

        if self.has(ssym):
            raise ValueError(
                'Time-domain expression has s.  Substitute with another symbol, say p')

        assumptions = self.assumptions.merge_and_infer(self, **assumptions)
        result = laplace_transform(
            self.expr, self.var, ssym, evaluate=evaluate,
            zero_initial_conditions=zero_initial_conditions)
        return self.change(result, domain='laplace', units_scale=uu.s, **assumptions)

    def laplace(self, evaluate=True, zero_initial_conditions=None,
                **assumptions):
        """Determine one-sided Laplace transform with 0- as the lower limit.

        By default initial conditions are assumed to be zero.  This
        can be controlled by `zero_initial_conditions`.

        This is an alias for LT."""

        return self.LT(evaluate=evaluate, zero_initial_conditions=zero_initial_conditions,
                       **assumptions)

    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        from .phasor import PhasorDomainExpression

        return PhasorDomainExpression.from_time(self, **assumptions)

    def FT(self, var=None, evaluate=True, **assumptions):
        """Attempt Fourier transform.

        X(f) = \int_{-\infty} ^ {\infty} x(t) exp(-j 2\pi f t) dt."""

        from .symbols import f, omega, Omega, F

        if var is None:
            var = f
        if id(var) not in (id(f), id(F), id(omega), id(Omega)):
            raise ValueError(
                'FT requires var to be f, F, omega, or Omega`, not %s' % var)

        if self.has(var):
            raise ValueError(
                'Time-domain expression has %s.  Substitute with another symbol, say %s0' % (var, var))

        assumptions = self.assumptions.merge_and_infer(self, **assumptions)
        result = fourier_transform(
            self.expr, self.var, fsym, evaluate=evaluate)
        result = self.change(result, domain='fourier',
                             units_scale=uu.s, **assumptions)
        result = result(var)
        result = result.expand(diracdelta=True, wrt=var)
        result = result.simplify()
        return result

    def fourier(self, var=None, evaluate=True, **assumptions):
        """Attempt Fourier transform. This is an alias for FT."""

        return self.FT(var, evaluate, **assumptions)

    def norm_fourier(self, evaluate=True, **assumptions):
        """Attempt normalized Fourier transform."""

        from .symbols import F

        return self.FT(F, evaluate, **assumptions)

    def angular_fourier(self, evaluate=True, **assumptions):
        """Attempt angular Fourier transform."""

        from .symbols import omega

        return self.FT(omega, evaluate, **assumptions)

    def norm_angular_fourier(self, evaluate=True, **assumptions):
        """Attempt normalized angular Fourier transform."""

        from .symbols import Omega

        return self.FT(Omega, evaluate, **assumptions)

    def frequency_response(self, **assumptions):
        """Convert to frequency response domain.  Note, this is similar to the
        Fourier domain but not always."""

        return self.laplace(**assumptions).frequency_response()

    def angular_frequency_response(self, **assumptions):
        """Convert to angular frequency response domain.  Note, this is
        similar to the angular Fourier domain but not always."""

        return self.laplace(**assumptions).angular_frequency_response()

    def time(self, **assumptions):
        return self

    def plot(self, t=None, **kwargs):
        """Plot the time waveform.  If t is not specified, it defaults to the
        range(-0.2, 2).  t can be a vector of specified instants, a
        tuple specifing the range, or a constant specifying the
        maximum value with the minimum value set to 0.

        kwargs include:
        `axes` - the plot axes to use otherwise a new figure is created
        `xlabel` - the x-axis label
        `ylabel` - the y-axis label
        `xscale` - the x-axis scaling, say for plotting as ms
        `yscale` - the y-axis scaling, say for plotting mV
        `plot_deltas` - plot Dirac deltas as arrows
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned."""

        from .plot import plot_time
        return plot_time(self, t, **kwargs)

    def response(self, xvector, tvector, method='bilinear'):
        """Evaluate response to input signal `xvector` at times
        `tvector`.  This returns a NumPy array."""

        from .sexpr import s
        return self(s).response(xvector, tvector, method=method)

    def sample(self, tvector):
        """Return a discrete-time signal evaluated at time values
        specified by `tvector`.  This returns a NumPy array."""

        return self.evaluate(tvector)

    def initial_value(self):
        """Determine value at t = 0.
        See also pre_initial_value and post_initial_value"""

        return self.subs(0)

    def pre_initial_value(self):
        """Determine value at t = 0-.
        See also initial_value and post_initial_value"""

        return self.limit(self.var, 0, dir='-')

    def post_initial_value(self):
        """Determine value at t = 0+.
        See also pre_initial_value and initial_value"""

        return self.limit(self.var, 0, dir='+')

    def final_value(self):
        """Determine value at t = oo."""

        return self.limit(self.var, oo)

    def force_causal(self):
        """Remove the piecewise condition from the expression
        and multiply by Heaviside function.  See also remove_condition."""

        if self.is_causal:
            return self

        expr = self.expr
        if self.is_conditional:
            expr = expr.args[0].args[0]
        expr = expr * Heaviside(t.var)
        return self.__class__(expr)

    def discretize(self, method=None, alpha=0.5):
        """Convert to a discrete-time approximation:

        :math:`x(n) \approx K x(t)`

        where :math:`K` is a scale factor.

        The default method depends on the expression quantity.  If the
        quantity is `undefined`, `voltage`, `current,
        `voltagesquared`, `currentsquared`, or `power`, the default
        method is `impulse-invariance` and there is no scaling.  This
        uses:

        : math: `x[n] = x(n \Delta t)`

        For other quantities, the default method is `bilinear', and
        the result is scaled by the sampling interval (`Delta t`) as
        is common for the discrete-time impulse response of digital
        filters:

        : math: `h[n] = \Delta t h_c(n \Delta t)`

        The methods are: 'impulse-invariance' 'bilinear', 'tustin',
        'trapezoidal' 'generalized-bilinear', 'gbf' controlled by the
        parameter `alpha` 'euler', 'forward-diff', 'forward-euler'
        'backward-diff', 'backward-euler' 'simpson', 'matched-Z',
        'zero-pole-matching'.  All the methods except
        `impulse-invariance` are applied in the Laplace domain so the
        result is unknown for `n < 0` unless the expression is causal.

        """

        from .sym import dt
        from .symbols import n

        signal = self.is_signal or self.is_squared or \
            self.is_power or self.is_undefined

        if method is None:
            if signal:
                method = 'impulse-invariance'
            else:
                method = 'bilinear'

        if method in ('impulse-invariance', ):
            from .symbols import t

            result = self.subs(t, n * dt)
            if not signal:
                result *= dt
            return result

        Hc = self.LT()
        H = Hc.discretize(method=method, alpha=alpha)
        if self.is_undefined:
            H /= dt
        return H.IZT()

    def dlti_filter(self, method='bilinear'):
        """Create DLTI filter using bilinear transform."""

        return self.LT().dlti_filter(method)

    def zdomain(self, **assumptions):

        # Going via the Laplace domain will restrict result for n >= 0.
        # Perhaps just sample with t = n * dt and warn if expression
        # contains Dirac deltas?
        return self.laplace().discretize(**assumptions)

    def discrete_frequency(self, **assumptions):

        return self.discrete_time(**assumptions).discrete_frequency()

    def discrete_time(self, **assumptions):

        return self.discretize(**assumptions)

    def differential_equation(self, inputsym='x', outputsym='y'):
        """Create differential equation from impulse response.
        """

        H = self.LT(zero_initial_conditions=True)
        return H.differential_equation(inputsym, outputsym)


class TimeDomainImpulseResponse(TimeDomainExpression):
    """Time-domain impulse response."""

    # TODO, check attributes.
    quantity = 'transfer'
    quantity_label = 'Impulse response'
    domain_units = uu.Hz
    is_transfer = True


def texpr(arg, **assumptions):
    """Create TimeDomainExpression object.  If `arg` is tsym return t"""

    if arg is tsym:
        return t
    return expr_make('time', arg, **assumptions)


t = TimeDomainExpression('t')
t.units = uu.s



"""This module provides the UndefinedDomainExpression class.

Copyright 2022 Michael Hayes, UCECE

"""

from .expr import Expr, expr_make
from .domains import UndefinedDomain
from .sym import usersymbol
from warnings import warn

__all__ = ('UndefinedDomainExpression', 'uexpr', )


class UndefinedDomainExpression(UndefinedDomain, Expr):

    def __init__(self, val, var=None, **assumptions):

        if var is None and isinstance(val, Expr):
            var = val.var
        if var is None:
            raise ValueError('Unspecified independent variable')

        self.var = usersymbol(var, override=False)
        assumptions['var'] = self.var

        super(UndefinedDomainExpression, self).__init__(val, **assumptions)

        if not self.has(self.var):
            warn('Expression %s does not contain variable %s' % (self, self.var))

    def plot(self, x=None, xlabel='', **kwargs):
        """Plot the expression.  If x is not specified, it defaults to the
        range(-0.2, 2).  x can be a vector of specified instants, a
        tuple specifing the range, or a constant specifying the
        maximum value with the minimum value set to 0.

        kwargs include:
        `axes` - the plot axes to use otherwise a new figure is created
        `xlabel` - the x-axis label
        `ylabel` - the y-axis label
        `xscale` - the x-axis scaling, say for plotting as ms
        `yscale` - the y-axis scaling, say for plotting mV
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned."""

        from .plot import plot_time
        return plot_time(self, x, xlabel=xlabel, **kwargs)


def uexpr(arg, var, **assumptions):
    """Create UndefinedDomainExpression object where `var` is the
    independent variable."""

    assumptions['var'] = var
    return expr_make('undefined', arg, **assumptions)



from lcapy import *
import unittest
from sympy import Rational


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_LSection(self):
        """Check LSection

        """
        a = LSection(R(10), R(30))

        self.assertEqual(a.Vgain12, 0.75, "Vgain12 incorrect.")
        self.assertEqual(a.Igain12, -1, "Igain12 incorrect.")
        self.assertEqual(a.Z1oc, impedance(40), "Z1oc incorrect.")
        self.assertEqual(a.Z2oc, impedance(30), "Z2oc incorrect.")
        self.assertEqual(a.Z1sc, impedance(10), "Z1sc incorrect.")
        self.assertEqual(a.Z2sc, impedance(7.5), "Z2sc incorrect.")
        self.assertEqual(a.Vgain(1, 2), 0.75, "Vgain 1->2 incorrect.")
        self.assertEqual(a.Vgain(2, 1), 1, "Vgain 2->1 incorrect.")
        self.assertEqual(a.Igain(1, 2), -1, "Igain 1->2 incorrect.")
        self.assertEqual(a.Igain(2, 1), -0.75, "Igain 2->1 incorrect.")
        self.assertEqual(
            a.Vresponse(Vdc(1), 1, 2), voltage(0.75 / s), "Vresponse 1->2 incorrect.")
        self.assertEqual(
            a.Vresponse(Vdc(1), 2, 1), voltage(1 / s), "Vresponse 2->1 incorrect.")
        self.assertEqual(
            a.Iresponse(Idc(1), 1, 2), current(-1 / s), "Iresponse 1->2 incorrect.")
        self.assertEqual(
            a.Iresponse(Idc(1), 2, 1), current(-0.75 / s), "Iresponse 2->1 incorrect.")

        b = a.load(R(30))
        self.assertEqual(b.Z, impedance(25), "R loaded Z incorrect.")
        self.assertEqual(b.Voc.dc, 0, "R loaded V incorrect.")

        c = a.load(R(30) + Vdc(5))
        self.assertEqual(c.Z, impedance(25), "T loaded Z incorrect.")
        self.assertEqual(c.Voc(s), voltage(2.5 / s), "T loaded V incorrect.")

        d = LSection(R(10), R(30) + Vdc(6))
        self.assertEqual(d.Vgain12, 0.75, "Vgain12 incorrect.")
        self.assertEqual(d.Igain12, -1, "Igain12 incorrect.")
        self.assertEqual(d.Z1oc, impedance(40), "Z1oc incorrect.")
        self.assertEqual(d.Z2oc, impedance(30), "Z2oc incorrect.")
        self.assertEqual(d.V1oc, voltage(6 / s), "V1oc incorrect.")
        self.assertEqual(d.V2oc, voltage(6 / s), "V2oc incorrect.")
        self.assertEqual(d.I1sc, 0 / s, "I2sc incorrect.")
        self.assertEqual(d.I2sc, current(-0.2 / s), "I2sc incorrect.")

        e = d.chain(Series(R(10)))
        self.assertEqual(e.Z1oc, impedance(40), "Z1oc incorrect.")
        self.assertEqual(e.Z2oc, impedance(40), "Z2oc incorrect.")
        self.assertEqual(e.Vgain12, 0.75, "Vgain12 incorrect.")
        self.assertEqual(e.Igain12, -0.75, "Igain12 incorrect.")

    def test_Shunt_parallel(self):
        """Check Shunts in parallel

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.chain(Shunt(R(30)))
        self.assertEqual(b.Z1oc, impedance(7.5), "Z1oc incorrect.")
        self.assertEqual(b.Z2oc, impedance(7.5), "Z2oc incorrect.")
        self.assertEqual(b.V1oc, voltage(3.75 / s), "V1oc incorrect.")
        self.assertEqual(b.V2oc, voltage(3.75 / s), "V2oc incorrect.")

    def test_Series_series(self):
        """Check Series in series

        """

        a = Series(R(10) + Vdc(5))
        b = a.chain(Series(R(30)))
        self.assertEqual(b.Y1sc, admittance(0.025), "Y1sc incorrect.")
        self.assertEqual(b.Y2sc, admittance(0.025), "Y2sc incorrect.")
        # This needs some thinking...
        # self.assertEqual(b.I1sc, -0.125 / s, "I1sc incorrect.")
        # self.assertEqual(b.I2sc, 0.125 / s, "I2sc incorrect.")

        c = a.chain(a)

    def test_LSection_models(self):
        """Check LSection models

        """
        a = LSection(R(10) + Vdc(5), R(20))

        self.assertEqual(a.Bparams.B11, 1, "incorrect B11.")
        self.assertEqual(a.Bparams.B12, -10, "incorrect B12.")
        # Note, -1 / 20 = -0.5 cannot be represented exactly as a float
        self.assertEqual(a.Bparams.B21, Rational(-1) / 20, "incorrect B21.")
        self.assertEqual(a.Bparams.B22, 1.5, "incorrect B22.")

        b = a.Zmodel

        self.assertEqual(b.Bparams.B11, 1, "incorrect B11.")
        self.assertEqual(b.Bparams.B12, -10, "incorrect B12.")
        self.assertEqual(b.Bparams.B21, Rational(-1) / 20, "incorrect B21.")
        self.assertEqual(b.Bparams.B22, 1.5, "incorrect B22.")

        c = a.Ymodel

        self.assertEqual(c.Bparams.B11, 1, "incorrect B11.")
        self.assertEqual(c.Bparams.B12, -10, "incorrect B12.")
        self.assertEqual(c.Bparams.B21, Rational(-1) / 20, "incorrect B21.")
        self.assertEqual(c.Bparams.B22, 1.5, "incorrect B22.")

        d = a.Hmodel

        self.assertEqual(d.Bparams.B11, 1, "incorrect B11.")
        self.assertEqual(d.Bparams.B12, -10, "incorrect B12.")
        self.assertEqual(d.Bparams.B21, Rational(-1) / 20, "incorrect B21.")
        self.assertEqual(d.Bparams.B22, 1.5, "incorrect B22.")

    def test_load(self):
        """Check load

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.load(R(30))

        self.assertEqual(b.Z, impedance(7.5), "Shunt loaded R incorrect Z.")

    def test_open_circuit(self):
        """Check open_circuit

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.open_circuit()

        self.assertEqual(b.Z, impedance(10), "incorrect Z.")
        self.assertEqual(b.Voc(s), voltage(5 / s), "incorrect V.")

    def test_short_circuit(self):
        """Check short_circuit

        """

        a = Series(R(10) + Vdc(5))
        b = a.short_circuit()

        self.assertEqual(b.Z, impedance(10), "incorrect Z.")
        self.assertEqual(b.Voc(s), voltage(5 / s), "incorrect V.")

    def test_Shunt(self):
        """Check Shunt

        """
        a = Shunt(R(10) + Vdc(5))
        self.assertEqual(a.Z1oc, impedance(10), "Z1oc incorrect.")
        self.assertEqual(a.Z2oc, impedance(10), "Z2oc incorrect.")
        self.assertEqual(a.V2oc, voltage(5 / s), "V2oc incorrect.")
        self.assertEqual(a.V1oc, voltage(5 / s), "V1oc incorrect.")
        # Cannot determine I1sc, I2sc, Ymn

        b = a.chain(Shunt(R(30)))
        self.assertEqual(b.Z1oc, impedance(7.5), "Z1oc incorrect.")
        self.assertEqual(b.Z2oc, impedance(7.5), "Z2oc incorrect.")
        self.assertEqual(b.V2oc, voltage(3.75 / s), "V2oc incorrect.")
        self.assertEqual(b.V1oc, voltage(3.75 / s), "V1oc incorrect.")

        c = a.chain(Series(R(30)))
        self.assertEqual(c.Z1oc, impedance(10), "Z1oc incorrect.")
        self.assertEqual(c.Z2oc, impedance(40), "Z2oc incorrect.")
        self.assertEqual(c.V1oc, voltage(5 / s), "V1oc incorrect.")

        self.assertEqual(c.V2oc, voltage(5 / s), "V2oc incorrect.")
        self.assertEqual(c.I1sc, current(-0.5 / s), "I1sc incorrect.")
        self.assertEqual(c.I2sc, current(0 / s), "I2sc incorrect.")

        d = Shunt(R(10)).chain(Series(R(30)))

    def test_Series(self):
        """Check Series

        """
        a = Series(R(10) + Vdc(5))
        self.assertEqual(a.Y1sc, admittance(0.1), "Y1sc incorrect.")
        self.assertEqual(a.Y2sc, admittance(0.1), "Y2sc incorrect.")
        self.assertEqual(a.I1sc, current(0.5 / s), "I1sc incorrect.")
        self.assertEqual(a.I2sc, current(-0.5 / s), "I2sc incorrect.")
        # Cannot determine V1oc, V2oc, Zmn

    def test_LSection(self):
        """Check LSection

        """
        a = LSection(R(10), R(30))

        self.assertEqual(a.Vgain12, 0.75, "Vgain12 incorrect.")
        self.assertEqual(a.Igain12, -1, "Igain12 incorrect.")
        self.assertEqual(a.Z1oc, impedance(40), "Z1oc incorrect.")
        self.assertEqual(a.Z2oc, impedance(30), "Z2oc incorrect.")
        self.assertEqual(a.Z1sc, impedance(10), "Z1sc incorrect.")
        self.assertEqual(a.Z2sc, impedance(7.5), "Z2sc incorrect.")
        self.assertEqual(a.Vgain(1, 2), 0.75, "Vgain 1->2 incorrect.")
        self.assertEqual(a.Vgain(2, 1), 1, "Vgain 2->1 incorrect.")
        self.assertEqual(a.Igain(1, 2), -1, "Igain 1->2 incorrect.")
        self.assertEqual(a.Igain(2, 1), -0.75, "Igain 2->1 incorrect.")
        self.assertEqual(
            a.Vresponse(Vdc(1), 1, 2), voltage(0.75 / s), "Vresponse 1->2 incorrect.")
        self.assertEqual(
            a.Vresponse(Vdc(1), 2, 1), voltage(1 / s), "Vresponse 2->1 incorrect.")
        self.assertEqual(
            a.Iresponse(Idc(1), 1, 2), current(-1 / s), "Iresponse 1->2 incorrect.")
        self.assertEqual(
            a.Iresponse(Idc(1), 2, 1), current(-0.75 / s), "Iresponse 2->1 incorrect.")

        b = a.load(R(30))
        self.assertEqual(b.Z, impedance(25), "R loaded Z incorrect.")
        self.assertEqual(b.Voc.dc, 0, "R loaded V incorrect.")

        c = a.load(R(30) + Vdc(5))
        self.assertEqual(c.Z, impedance(25), "T loaded Z incorrect.")
        self.assertEqual(c.Voc(s), voltage(2.5 / s), "T loaded V incorrect.")

        d = LSection(R(10), R(30) + Vdc(6))
        self.assertEqual(d.Vgain12, 0.75, "Vgain12 incorrect.")
        self.assertEqual(d.Igain12, -1, "Igain12 incorrect.")
        self.assertEqual(d.Z1oc, impedance(40), "Z1oc incorrect.")
        self.assertEqual(d.Z2oc, impedance(30), "Z2oc incorrect.")
        self.assertEqual(d.V1oc, voltage(6 / s), "V1oc incorrect.")
        self.assertEqual(d.V2oc, voltage(6 / s), "V2oc incorrect.")
        self.assertEqual(d.I1sc, current(0 / s), "I2sc incorrect.")
        self.assertEqual(d.I2sc, current(-0.2 / s), "I2sc incorrect.")

        e = d.chain(Series(R(10)))
        self.assertEqual(e.Z1oc, impedance(40), "Z1oc incorrect.")
        self.assertEqual(e.Z2oc, impedance(40), "Z2oc incorrect.")
        self.assertEqual(e.Vgain12, 0.75, "Vgain12 incorrect.")
        self.assertEqual(e.Igain12, -0.75, "Igain12 incorrect.")

    def test_Shunt_parallel(self):
        """Check Shunts in parallel

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.chain(Shunt(R(30)))
        self.assertEqual(b.Z1oc, impedance(7.5), "Z1oc incorrect.")
        self.assertEqual(b.Z2oc, impedance(7.5), "Z2oc incorrect.")
        self.assertEqual(b.V1oc, voltage(3.75 / s), "V1oc incorrect.")
        self.assertEqual(b.V2oc, voltage(3.75 / s), "V2oc incorrect.")

    def test_Series_series(self):
        """Check Series in series

        """

        a = Series(R(10) + Vdc(5))
        b = a.chain(Series(R(30)))
        self.assertEqual(b.Y1sc, admittance(0.025), "Y1sc incorrect.")
        self.assertEqual(b.Y2sc, admittance(0.025), "Y2sc incorrect.")
        # The following fail and need some thinking...
        # self.assertEqual(b.I1sc, -0.125 / s, "I1sc incorrect.")
        # self.assertEqual(b.I2sc, 0.125 / s, "I2sc incorrect.")

    def test_load(self):
        """Check load

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.load(R(30))

        self.assertEqual(b.Z, impedance(7.5), "Shunt loaded R incorrect Z.")

    def test_open_circuit(self):
        """Check open_circuit

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.open_circuit()

        self.assertEqual(b.Z, impedance(10), "incorrect Z.")
        self.assertEqual(b.Voc(s), voltage(5 / s), "incorrect V.")

    def test_short_circuit(self):
        """Check short_circuit

        """

        a = Series(R(10) + Vdc(5))
        b = a.short_circuit()

        self.assertEqual(b.Z, impedance(10), "incorrect Z.")
        self.assertEqual(b.Voc(s), voltage(5 / s), "incorrect V.")

    def test_LSection_models(self):
        """Check LSection models

        """
        a = LSection(R(10) + Vdc(5), R(20))

        self.assertEqual(a.Bparams.B11, 1, "incorrect B11.")
        self.assertEqual(a.Bparams.B12, -10, "incorrect B12.")
        self.assertEqual(a.Bparams.B21, Rational(-1) / 20, "incorrect B21.")
        self.assertEqual(a.Bparams.B22, 1.5, "incorrect B22.")

        b = a.Zmodel

        self.assertEqual(b.Bparams.B11, 1, "incorrect B11.")
        self.assertEqual(b.Bparams.B12, -10, "incorrect B12.")
        self.assertEqual(b.Bparams.B21, Rational(-1) / 20, "incorrect B21.")
        self.assertEqual(b.Bparams.B22, 1.5, "incorrect B22.")

        c = a.Ymodel

        self.assertEqual(c.Bparams.B11, 1, "incorrect B11.")
        self.assertEqual(c.Bparams.B12, -10, "incorrect B12.")
        self.assertEqual(c.Bparams.B21, Rational(-1) / 20, "incorrect B21.")
        self.assertEqual(c.Bparams.B22, 1.5, "incorrect B22.")

        d = a.Hmodel

        self.assertEqual(d.Bparams.B11, 1, "incorrect B11.")
        self.assertEqual(d.Bparams.B12, -10, "incorrect B12.")
        self.assertEqual(d.Bparams.B21, Rational(-1) / 20, "incorrect B21.")
        self.assertEqual(d.Bparams.B22, 1.5, "incorrect B22.")

    def test_transforms(self):

        A = AMatrix.generic()
        self.assertEqual(A.Aparams.Aparams.simplify(), A, "A.Aparams.Aparams")
        self.assertEqual(A.Bparams.Aparams.simplify(), A, "A.Bparams.Aparams")
        self.assertEqual(A.Gparams.Aparams.simplify(), A, "A.Gparams.Aparams")
        self.assertEqual(A.Hparams.Aparams.simplify(), A, "A.Hparams.Aparams")
        self.assertEqual(A.Sparams.Aparams.simplify(), A, "A.Sparams.Aparams")
        self.assertEqual(A.Tparams.Aparams.simplify(), A, "A.Tparams.Aparams")
        self.assertEqual(A.Yparams.Aparams.simplify(), A, "A.Yparams.Aparams")
        self.assertEqual(A.Zparams.Aparams.simplify(), A, "A.Zparams.Aparams")

        Z = ZMatrix.generic()
        self.assertEqual(Z.Aparams.Zparams.simplify(), Z, "Z.Aparams.Zparams")
        self.assertEqual(Z.Bparams.Zparams.simplify(), Z, "Z.Bparams.Zparams")
        self.assertEqual(Z.Gparams.Zparams.simplify(), Z, "Z.Gparams.Zparams")
        self.assertEqual(Z.Hparams.Zparams.simplify(), Z, "Z.Hparams.Zparams")
        self.assertEqual(Z.Sparams.Zparams.simplify(), Z, "Z.Sparams.Zparams")
        self.assertEqual(Z.Tparams.Zparams.simplify(), Z, "Z.Tparams.Zparams")
        self.assertEqual(Z.Yparams.Zparams.simplify(), Z, "Z.Yparams.Zparams")
        self.assertEqual(Z.Zparams.Zparams.simplify(), Z, "Z.Zparams.Zparams")

    def test_twoport_print(self):

        a = Series(C(1)).chain(Shunt(L(2))).chain(Series(R(3)))
        self.assertEqual(
            a.pretty(), 'Chain(Chain(Series(C(1)), Shunt(L(2))), Series(R(3)))', 'pretty')
        self.assertEqual(
            str(a), 'Chain(Chain(Series(C(1)), Shunt(L(2))), Series(R(3)))', 'str')
        self.assertEqual(
            repr(a), 'Chain(Chain(Series(C(1)), Shunt(L(2))), Series(R(3)))', 'repr')
        self.assertEqual(a.latex(
        ), '\\mathrm{Chain}(\\mathrm{Chain}(\\mathrm{Series}(\\mathrm{C}(1)), \\mathrm{Shunt}(\\mathrm{L}(2))), \\mathrm{Series}(\\mathrm{R}(3)))', 'latex')

    def test_params(self):

        attrs = ('Z1sc', 'Z1oc', 'Z2sc', 'Z2oc', 'Vgain12',
                 'Vgain21', 'Igain12', 'Igain21')

        models = ('Aparams', 'Bparams', 'Gparams',
                  'Hparams', 'Yparams', 'Zparams')

        a = TSection(R(1), R(2), R(3))

        vals = {}

        for attr in attrs:
            vals[attr] = getattr(a, attr)

        for model in models:
            params = getattr(a, model)
            for attr in attrs:
                self.assertEqual(vals[attr], getattr(
                    params, attr), 'Mismatch for %s.%s' % (model, attr))

    def test_circuit_to_ladder(self):

        a = Circuit("""
P1 1 2; down=1.5
W1 1 3; right
C1 3 4; down=1.5
W2 4 2; left
R1 3 5; right=1.5
W3 4 6; right=1.5
C2 5 6; down=1.5
R2 5 7; right=1.5
W4 6 8; right=1.5
P2 7 8; down=1.5
""")
        l = a.ladder('P1', 'P2')
        self.assertEqual(l.args[0].C, C('C1').C, 'args[0]')

        Vgain1 = l.voltage_gain.simplify()
        Vgain = expr('1 / (C2 * R1 * s + 1)')

        self.assertEqual(Vgain, Vgain1, 'voltage gain')

    def test_circuit_to_alt_ladder(self):

        a = Circuit("""
R1 3 5; right=1.5
W3 4 6; right=1.5
R2 5 7; right=1.5
W4 6 8; right=1.5
P2 7 8; down=1.5
P1 3 4; down=1.5
C1 5 6; down=1.5
""")
        l = a.ladder('P1', 'P2')
        self.assertEqual(l.args[0].R, R('R1').R, 'args[0]')

        Vgain1 = l.voltage_gain.simplify()
        Vgain = expr('1 / (C1 * R1 * s + 1)')

        self.assertEqual(Vgain, Vgain1, 'voltage gain')



"""This module implements the LaplaceDomainMatrix class for a matrix of
Laplace-domain expressions.

Copyright 2019--2023 Michael Hayes, UCECE

"""

from .matrix import Matrix


class LaplaceDomainMatrix(Matrix):
    from .exprclasses import LaplaceDomainExpression
    _typewrap = LaplaceDomainExpression

    def ILT(self, **assumptions):

        def func(expr):
            return expr.ILT(**assumptions)

        return TimeDomainMatrix(self.applyfunc(func))


class LaplaceDomainVoltageMatrix(LaplaceDomainMatrix):
    from .exprclasses import LaplaceDomainVoltage
    _typewrap = LaplaceDomainVoltage


class LaplaceDomainCurrentMatrix(LaplaceDomainMatrix):
    from .exprclasses import LaplaceDomainCurrent
    _typewrap = LaplaceDomainCurrent


class LaplaceDomainAdmittanceMatrix(LaplaceDomainMatrix):
    from .exprclasses import LaplaceDomainAdmittance
    _typewrap = LaplaceDomainAdmittance


class LaplaceDomainImpedanceMatrix(LaplaceDomainMatrix):
    from .exprclasses import LaplaceDomainImpedance
    _typewrap = LaplaceDomainImpedance


from .tmatrix import TimeDomainMatrix  # nopep8



"""
This module implements the Vector class (a wrapper for Matrix).

Copyright 2020--2021 Michael Hayes, UCECE
"""

from __future__ import division
from .matrix import Matrix
from .sym import sympify
from .expr import expr


class Vector(Matrix):

    def __new__(cls, *args, **assumptions):

        if len(args) == 2:
            return super(Vector, cls).__new__(cls, (expr(args[0], **assumptions).expr, expr(args[1], **assumptions).expr))

        args = [expr(arg, **assumptions).sympy for arg in args[0]]

        return super(Vector, cls).__new__(cls, args)



class SeqDomain(object):

    is_sequence = True

    def _class_get(self, quantity, domain):

        from .seqmap import seqmap
        return seqmap(quantity, domain)

    def _class_by_quantity(self, quantity, domain=None):

        if domain is None:
            domain = self.domain
        return self._class_get(quantity, domain)

    def _class_by_domain(self, domain):

        return self._class_get(self.quantity, domain)

    def as_quantity(self, quantity):

        if quantity == 'voltage':
            return self.as_voltage()
        elif quantity == 'current':
            return self.as_current()
        elif quantity == 'impedance':
            return self.as_impedance()
        elif quantity == 'admittance':
            return self.as_admittance()
        elif quantity == 'transfer':
            return self.as_transfer()
        elif quantity == 'power':
            return self.as_power()
        elif quantity == 'undefined':
            return self.as_expr()
        raise ValueError('Unknown quantity %s for %s' % (quantity, self))

    def as_domain(self, domain):

        if domain == 'discrete time':
            return self.as_discrete_time()
        elif domain == 'discrete fourier':
            return self.as_discrete_fourier()
        elif domain == 'Z':
            return self.as_Z()
        raise ValueError('Unknown domain %s for %s' % (domain, self))

    def as_voltage(self):
        return self._class_by_quantity('voltage')(self)

    def as_current(self):
        return self._class_by_quantity('current')(self)

    def as_admittance(self):
        return self._class_by_quantity('admittance')(self)

    def as_impedance(self):
        return self._class_by_quantity('impedance')(self)

    def as_transfer(self):
        return self._class_by_quantity('transfer')(self)

    def as_power(self):
        return self._class_by_quantity('power')(self)

    def change(self, arg, domain=None, units_scale=None, **assumptions):
        """Change sequence class."""

        if domain is None:
            domain = self.domain

        quantity = self.quantity

        cls = self._class_by_quantity(quantity, domain)
        ret = cls(arg, **assumptions)
        return ret



"""This module provides the NetlistMaker class for generating
horizontal or vertical netlists from a network.

Copyright 2020 Michael Hayes, UCECE

"""

from .netlisthelper import NetlistHelper

class NetlistMaker(NetlistHelper):

    def __init__(self, net, layout='horizontal', evalf=None):

        self.net = net
        self.evalf = evalf
        self.names = []

        if layout == 'horizontal':
            self.dir = 'right'
        elif layout == 'vertical':
            self.dir = 'down'
        else:
            raise ValueError('Unknown layout ' + layout)

    def __call__(self):

        n1 = self._node
        n2 = self._node
        return self.net._net_make(self, n2, n1, dir=self.dir)



from lcapy import impedance

Z = impedance('(4*s**2 + 3 * s + 1 / 6) / (s**2 + 2 * s / 3)')

n = Z.network('fosterI')

n.draw(__file__.replace('py', 'png'))
#n.draw(__file__.replace('py', 'png'), layout='ladder')



"""This module implements the TimeDomainMatrix class for a matrix of
time-domain expressions.

Copyright 2019--2021 Michael Hayes, UCECE

"""

from .matrix import Matrix


class TimeDomainMatrix(Matrix):
    from .texpr import TimeDomainExpression
    _typewrap = TimeDomainExpression

    def LT(self):

        def func(expr):
            return expr.LT()

        return LaplaceDomainMatrix(self.applyfunc(func))


from .smatrix import LaplaceDomainMatrix  # nopep8



from lcapy import *
from numpy import arange
import unittest


class LcapyTester(unittest.TestCase):
    """Unit tests for lcapy"""

    def test_estimate1(self):

        e = expr('a * exp(-t  / tau) * u(t)')

        tv = arange(10)

        params = {'a': 1, 'tau': 10}

        vv = e.subs(params).evaluate(tv)

        eparams = e.estimate(tv, vv, method='trf',
                             ranges={'a': (0, 10), 'tau': (1, 20)}).params

        assert abs(params['a'] - eparams['a']) < 1e-6 and \
            abs(params['tau'] - eparams['tau']) < 1e-6

        eparams = e.estimate(tv, vv, method='Nelder-Mead',
                             ranges={'a': (0, 10), 'tau': (1, 20)}).params

        assert abs(params['a'] - eparams['a']) < 1e-5 and \
            abs(params['tau'] - eparams['tau']) < 1e-4

        eparams = e.estimate(tv, vv, method='brute',
                             ranges={'a': (0, 10), 'tau': (1, 20)}).params

        assert abs(params['a'] - eparams['a']) < 1e-5 and \
            abs(params['tau'] - eparams['tau']) < 1e-4



"""This module handles sequences.

Copyright 2020--2023 Michael Hayes, UCECE

"""

from .expr import ExprList, expr
from .seqdomain import SeqDomain
from .utils import isiterable
from .assumptions import Assumptions

# Perhaps subclass numpy ndarray?  But then could not have symbolic
# elements in the sequence.


def parse_seq_str(s):

    if s.startswith('{'):
        if not s.endswith('}'):
            raise ValueError('Mismatched braces for %s' % s)
        s = s[1:-1]

    parts = s.split(',')
    N = len(parts)

    vals = []
    m0 = None
    for m, item in enumerate(parts):
        item = item.strip()
        if item.startswith('_'):
            if m0 is not None:
                raise ValueError('Cannot have multiple zero index indicators')
            m0 = m
            item = item[1:]

        val = expr(item)
        vals.append(val)

    if m0 is None:
        m0 = 0

    ni = range(-m0, N - m0)
    return vals, ni


class Sequence(ExprList, SeqDomain):

    var = None
    is_sequence = True
    quantity = 'undefined'

    def __init__(self, seq, ni=None, origin=None, evaluate=False, var=None,
                 start_trunc=False, end_trunc=False):
        """Sequences can be created from an tuple, list, or ndarray.

        See `seq()` to create a Sequence from a string.

        >>> a = Sequence((1, 2, 3))

        The sequence indices are specified with the optional `ni` argument.
        For example:

        >>> a = Sequence((1, 2, 3, 4), (-1, 0, 1, 2))

        If the `ni` argument is not specified, the sequence indices
        are enumerated from 0.

        The sequence indices can be found using the `n` attribute.
        This returns a list.

        >>> a = Sequence('{1, _2, 3, 4}').n
        [-1, 0, 1, 2]

        Sequences can be converted into discrete-time, discrete-frequency,
        z-domain sequences using call notation, for example::

        >>> a(z)

        `start_trunc` indicates that the start of the sequence was truncated
        `end_trunc` indicates that the end of the sequence was truncated

        `start_trunc` and `end_trunc` are not propagated if a sequence is
        modified.  They indicate that ellipsis should be printed to show
        the sequence has been truncated.

        """

        if isinstance(seq, str):
            seq, ni = parse_seq_str(seq)

        if not isiterable(seq):
            seq = (seq, )

        super(Sequence, self).__init__(seq, evaluate)

        if ni is not None and origin is not None:
            raise ValueError('Cannot specify both ni and origin')

        if origin is not None:
            ni = range(-origin, len(self) - origin)

        if ni is None:
            ni = range(len(seq))

        # Perhaps enforce contiguous sequences and just store the origin?
        # This will simplify sequence comparison.
        self.n = list(ni)

        # Determine if sequence truncated at start, end, or both.
        # Perhaps have separate classes for truncated sequences?
        self.start_trunc = start_trunc
        self.end_trunc = end_trunc
        # For symmetry with Expr
        self.assumptions = Assumptions()

    @property
    def vals(self):
        """Return the SymPy values as a list."""
        return list(self)

    @property
    def _pexpr(self):
        """Return expression for printing."""
        return self

    def __eq__(self, x):

        self._check_compatible(x)
        return self.vals == x.vals and self.n == x.n

    def _check_compatible(self, x):

        if not isinstance(x, Sequence):
            raise TypeError('Can only add a sequence to a sequence')
        if x.quantity != 'undefined' and self.quantity != 'undefined' and x.quantity != self.quantity:
            raise TypeError('Sequences have different quantities: %s and %s' % (
                self.quantity, x.quantity))

        if self.domain != x.domain:
            raise TypeError('Sequences have different domains: %s and %s' % (
                self.domain, x.domain))

    def __abs__(self):
        """Absolute value of each element."""

        raise ValueError(
            'abs operator not supported for %s: use .as_array()' % self.__class__.__name__)

    def __neg__(self):
        """Negation of each element."""

        raise ValueError(
            '- operator not supported for %s: use .as_array()' % self.__class__.__name__)

    def __add__(self, x):
        """Concatenate with sequence x."""

        self._check_compatible(x)
        if self.quantity == 'undefined':
            cls = x.__class__
        else:
            cls = self.__class__

        return cls(super(Sequence, self).__add__(x))

    def __mul__(self, x):
        """Replicate x times."""

        return self.__class__(super(Sequence, self).__mul__(x))

    def __lshift__(self, m):

        return self.delay(-m)

    def __rshift__(self, m):

        return self.delay(m)

    def __str__(self):

        items = []
        if self.start_trunc:
            items.append('...')

        for v1, n1 in zip(self, self.n):
            s = str(v1)

            if n1 == 0:
                s = '_' + s
            items.append(s)

        if self.end_trunc:
            items.append('...')

        return r'{%s}' % ', '.join(items)

    def __getitem__(self, n):
        """Note this returns the element with index matching n.
        This is not necessarily the nth element in the sequence."""

        # TODO, support slices, etc.
        try:
            nindex = list(self.n).index(n)
        except ValueError:
            return expr(0)

        return super(Sequence, self).__getitem__(nindex)

    @property
    def origin(self):
        """Return the element index for n == 0. This may raise a ValueError
        if the origin is not in the sequence."""

        return -min(self.n)

    @origin.setter
    def origin(self, origin):
        """Set the origin to `origin`."""

        from numpy import arange

        self.n = list(arange(-origin, len(self) - origin))

    def prune(self):
        """Remove zeros from ends of sequence.

        {0, 0, 1, 2, 3, 0} -> {1, 2, 3}"""

        vals = self.vals

        m1 = 0
        while vals[m1] == 0:
            m1 += 1

        m2 = len(vals) - 1
        if vals[m2] != 0:
            return self.__class__(vals[m1:], self.n[m1:])

        while vals[m2] == 0:
            m2 -= 1
        return self.__class__(vals[m1:m2 + 1], self.n[m1:m2 + 1])

    def zeropad(self, M):
        """Add M zeros to end of sequence:

        For example, with M = 3

        {1, 2, 3} -> {1, 2, 3, 0, 0, 0}"""

        vals = self.vals
        n = self.n

        zero = expr(0)
        for m in range(M):
            vals.append(zero)

        ni = self.n + list(range(self.n[-1] + 1, len(vals)))
        return self.__class__(vals, ni)

    def __str__(self):

        a = self.zeroextend()

        items = []
        if self.start_trunc:
            items.append('...')

        for v1, n1 in zip(a, a.n):
            try:
                s = v1.latex()
            except:
                s = str(v1)

            if n1 == 0:
                s = r'_%s' % v1
            items.append(s)

        if self.end_trunc:
            items.append('...')
        return r'{%s}' % ', '.join(items)

    def pretty(self, **kwargs):

        from .printing import pretty

        a = self.zeroextend()
        return pretty(a)

    def as_array(self):
        """Numerically evaluate and store as NumPy array."""

        from numpy import array, allclose

        # If, for some reason, a sequence can have elements that
        # depend on n...
        #vals = array([v1.evaluate(n1) for v1, n1 in zip(self, self.n)])
        vals = array([v1.cval for v1 in self])

        if allclose(vals.imag, 0.0):
            vals = vals.real

        return vals

    @property
    def expr(self):
        """Convert sequence to an Lcapy expression."""

        if self.var is None:
            raise ValueError('var not specified')

        return self.as_impulses(self.var)

    def as_impulses(self, var=None):
        """Convert to discrete-time signal in the form of
        a weighted sum of delayed impulses.  For example,
        {1, 2, 3} -> ui[n] + 2 * ui[n - 1] + 3 * ui[n - 2]"""

        from .functions import unitimpulse
        from sympy import Add

        if var is None:
            var = self.var
        if var is None:
            raise ValueError('var not specified')

        # This can reorder terms
        result = var * 0
        for v1, n1 in zip(self, self.n):
            result += v1.as_constant() * unitimpulse(var - n1)

        # but so does the unevaluated Add...
        # items = []
        # for v1, n1 in zip(self, self.n):
        #     items.append(v1 * unitimpulse(var.var - n1))
        #
        # expr = Add(*items, evaluate=False)
        # result = var.__class__(expr)

        return result

    def evaluate(self, ni=None):
        """Evaluate expression at sequence indices specified by `arg`.  `arg`
        may be a scalar or a vector.  The result is of type float or
        complex.  Zeroes are returned for indices outside the sequence
        extent.

        If `arg` is iterable, a NumPy array is returned.

        """
        from numpy import array, allclose

        if ni is None:
            return self.as_array()
        if isiterable(ni):
            vals = array([self(n1).cval for n1 in ni])

            if allclose(vals.imag, 0.0):
                vals = vals.real

            return vals
        else:
            val = self(ni).cval

            if allclose(val.imag, 0.0):
                val = val.real

            return val

    @property
    def extent(self):
        """Determine extent of the sequence.

        For example, Sequence([1, 1]).extent = 2
                     Sequence([1, 0, 1]).extent = 3
                     Sequence([0, 1, 0, 1]).extent = 3
        """

        from numpy import argwhere

        # Note, each element is an Expr.
        nz = [elt != 0 for elt in self]
        w = argwhere(nz)
        if len(w) == 0:
            return 0

        return (max(w) - min(w))[0] + 1

    def discrete_time_fourier_transform(self, var=None, **assumptions):
        """Convert to Fourier domain using discrete time Fourier transform."""
        return self.DTFT(var, **assumptions)

    def DTFT(self, var=None, **assumptions):
        """Convert to Fourier domain using discrete time Fourier transform."""

        return self.as_impulses().DTFT(var, **assumptions)

    def plot(self, ni=None, **kwargs):
        """Plot the sequence.  If `ni` is not specified, it defaults to the
        sequence indices.  `ni` can be a vector of specified sequence
        indices, a tuple specifing the range, or a constant specifying
        the maximum value with the minimum value set to 0.

        kwargs include:
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label
        ylabel - the y-axis label
        xscale - the x-axis scaling, say for plotting as ms
        yscale - the y-axis scaling, say for plotting mV
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned.

        """

        if ni is None:
            ni = self.n

        # This is not the most efficient way but plot routines expect
        # an Expr object.
        return self.as_impulses().plot(ni, **kwargs)

    def __call__(self, arg, **assumptions):
        """Convert sequence to n-domain or k-domain expression.
        For example, seq((1, 2, 3))(n)"""

        from .symbols import n, k, z
        from .sym import nsym, ksym, zsym

        if id(arg) == id(n) or arg == n:
            if self.var == nsym:
                return self.copy()
            elif self.var == ksym:
                return self.IDFT()
            return self.IZT()
        if id(arg) == id(k) or arg == k:
            if self.var == ksym:
                return self.copy()
            elif self.var == zsym:
                return self.IZT().DFT()
            return self.DFT()
        if id(arg) == id(z) or arg == z:
            if self.var == zsym:
                return self.copy()
            elif self.var == ksym:
                return self.IDFT().ZT()
            return self.ZT()

        # This is experimental and may be deprecated.
        return self[arg]

    def _repr_pretty_(self, p, cycle):
        """This is used by jupyter notebooks to display an expression using
        unicode.  It is also called by IPython when displaying an
        expression."""

        # Note, the method in ExprPrint is bypassed since list
        # has this methodx.

        from .printing import pretty
        p.text(self.pretty())

    def copy(self):
        return self.__class__(super(Sequence, self).copy(),
                              self.n)

    def lfilter(self, b=None, a=None):
        """Implement digital filter specified by a transfer function.  The
        transfer function is described by a vector `b` of coefficients
        for the numerator and an `a` vector of coefficients for the
        denominator.

        If you would like the response with initial conditions see
        `DTfilter.response()`.

        For a FIR filter a = [1]."""

        if b is None:
            b = []
        if a is None:
            a = [1]

        x = self.vals
        y = []

        a0 = a[0]

        for n, x1 in enumerate(x):
            y.append(expr(0))

            for m, b1 in enumerate(b):
                try:
                    y[-1] += b1 * x[n - m] / a0
                except:
                    pass

            yn = y[-1]
            for m, a1 in enumerate(a[1:]):
                try:
                    yn += a1 * y[-m - 2] / a0
                except:
                    pass
            y[-1] = yn

        return self.__class__(y, self.n)

    def convolve(self, h, mode='full'):
        """Convolve with h."""

        x = self
        h = Sequence(h)

        Lx = x.extent
        Lh = h.extent
        Ly = Lx + Lh - 1

        if mode == 'full':
            x = x.zeropad(Ly - Lx)
        elif mode == 'same':
            x = x.zeropad(max(Lx, Ly) - Lx)
        else:
            raise ValueError('Unknown mode ' + mode)

        return x.lfilter(h, a=[1])

    def delay(self, m=0):
        """Return a new sequence delayed by an integer number of samples `m`.
        If `m` is negative, the sequence is advanced."""

        from numpy import arange

        if m != int(m):
            raise ValueError('Non-integer delay %s' % m)

        origin = self.origin - m
        ni = list(arange(-origin, len(self) - origin))

        return self.__class__(self.vals, ni)

    def zeroextend(self):
        """Extend sequence by adding zeros so that the origin
        is included.  This is used for printing."""

        ni = self.n
        vals = self.vals
        if ni[0] > 0:
            vals = [0] * ni[0] + vals
            ni = range(0, ni[-1] + 1)
        elif ni[-1] < 0:
            vals = vals + [0] * -ni[-1]
            ni = range(ni[0], 1)

        return self.__class__(vals, ni,
                              start_trunc=self.start_trunc,
                              end_trunc=self.end_trunc)



from lcapy import Istep, R, L, C, t
from matplotlib.pyplot import figure, savefig
import numpy as np

a = Istep(10) | R(0.1) | C(0.4) | L(0.2)

vt = np.linspace(0, 10, 1000)

fig = figure()
ax = fig.add_subplot(111)
ax.plot(vt, a.Voc(t).evaluate(vt), linewidth=2)
ax.set_xlabel('Time (s)')
ax.set_ylabel('Voltage (V)')
ax.grid(True)

savefig('parallel-IRLC1-voc.png')



"""
This module supports simple linear three-port networks.  It is
experimental and needs a rethink.

Copyright 2014--2021 Michael Hayes, UCECE

"""

from __future__ import division
from warnings import warn
import sympy as sym
from .sexpr import LaplaceDomainVoltage, LaplaceDomainTransferFunction
from .sexpr import LaplaceDomainCurrent
from .smatrix import LaplaceDomainVoltageMatrix, LaplaceDomainCurrentMatrix
from .smatrix import LaplaceDomainImpedanceMatrix, LaplaceDomainAdmittanceMatrix
from .cexpr import cexpr
from .oneport import OnePort
from .twoport import YMatrix, ZMatrix, TwoPortZModel, Series, TwoPort

__all__ = ('Opamp', )


class ThreePortMatrix(sym.Matrix):

    def __new__(cls, *args):

        if len(args) == 9:
            args = ((args[0], args[1], args[2]),
                    (args[3], args[4], args[5]),
                    (args[6], args[7], args[8]))
            return super(ThreePortMatrix, cls).__new__(cls, args)

        return super(ThreePortMatrix, cls).__new__(cls, *args)

    @property
    def Z(self):
        return ZMatrix3(self.Y.inv())

    @property
    def Y(self):
        return YMatrix3(self.Z.inv())


class ZMatrix3(ThreePortMatrix):

    """
    +-  -+     +-             -+   +-  -+
    | V1 |     | Z11  Z12  Z13 |   | I1 |
    | V2 |  =  | Z21  Z22  Z23 |   | I2 |
    | V3 |     | Z31  Z32  Z33 |   | I3 |
    +-  -+     +-             -+   +-  -+

    Z = inv(Y)
    """

    @property
    def Z(self):
        # Perhaps we should make a copy?
        return self


class YMatrix3(ThreePortMatrix):

    """
    +-  -+     +-             -+   +-  -+
    | I1 |  =  | Y11  Y12  Y13 |   | V1 |
    | I2 |     | Y21  Y22  Y23 |   | V2 |
    | I3 |     | Y31  Y32  Y33 |   | V3 |
    +-  -+     +-             -+   +-  -+

    Y = inv(Z)
    """

    @property
    def Y(self):
        # Perhaps we should make a copy?
        return self


class ThreePort(object):

    """

    +-  -+     +-             -+   +-  -+     +-   -+
    | V1 |     | Z11  Z12  Z13 |   | I1 |     | V1z |
    | V2 |  =  | Z21  Z22  Z23 |   | I2 |  +  | V2z |
    | V3 |     | Z31  Z32  Z33 |   | I3 |     | V3z |
    +-  -+     +-             -+   +-  -+     +-   -+

    The A, B, G, and H models are invalid for multiports.

    Unfortunately, the Z model can blow up for simple networks.

    """

    def __init__(self, Z, Vz=None):

        if Vz is None:
            Vz = LaplaceDomainVoltageMatrix((0, 0, 0))

        if not isinstance(Z, ZMatrix3):
            raise ValueError('Z not ZMatrix3')

        if not isinstance(Vz, LaplaceDomainVoltageMatrix):
            raise ValueError('Vz not LaplaceDomainVoltageMatrix')

        self._M = Z
        self._Vz = Vz

    @property
    def Voc(self):
        """Return voltage vector with all ports open-circuited
        (i.e., In = 0)"""
        return self._Vz

    @property
    def Isc(self):
        """Return current vector with all ports short-circuited
        (i.e., Vn = 0)"""
        Y = self.Y
        Voc = self.Voc

        Isc = LaplaceDomainCurrentMatrix(
            [Voc[m] * Y[m, m] for m in range(len(Voc))])
        return Isc

    @property
    def Y(self):
        """Return admittance matrix"""
        return YMatrix3(self._M.Y)

    @property
    def Z(self):
        """Return impedance matrix"""
        return self._M

    @property
    def Yoc(self):
        """Return admittance vector with ports open circuit"""
        Z = self.Z
        return LaplaceDomainAdmittanceMatrix([1 / Z[m, m] for m in range(Z.shape[0])])

    @property
    def Ysc(self):
        """Return admittance vector with ports short circuit"""
        Y = self.Y
        return LaplaceDomainAdmittanceMatrix([Y[m, m] for m in range(Y.shape[0])])

    @property
    def Zoc(self):
        """Return impedance vector with ports open circuit"""
        Z = self.Z
        return LaplaceDomainImpedanceMatrix([Z[m, m] for m in range(Z.shape[0])])

    @property
    def Zsc(self):
        """Return impedance vector with ports short circuit"""
        Y = self.Y
        return LaplaceDomainImpedanceMatrix([1 / Y[m, m] for m in range(Y.shape[0])])

    def _portcheck(self, port):

        if port not in (1, 2, 3):
            raise ValueError('Invalid port ' + port)

    def Vgain(self, inport=1, outport=2):
        """Return voltage gain for specified ports with internal
        sources zero"""

        self._portcheck(inport)
        self._portcheck(outport)

        p1 = inport - 1
        p2 = outport - 1

        return LaplaceDomainTransferFunction(self.Z[p2, p1] / self.Z[p1, p1])

    def Igain(self, inport=1, outport=2):
        """Return voltage gain for specified ports with internal
        sources zero"""

        self._portcheck(inport)
        self._portcheck(outport)

        p1 = inport - 1
        p2 = outport - 1

        return LaplaceDomainTransferFunction(self.Y[p2, p1] / self.Y[p1, p1])

    def Vresponse(self, V, inport=1, outport=2):
        """Return voltage response for specified applied voltage and
        specified ports"""

        self._portcheck(inport)
        self._portcheck(outport)

        p1 = inport - 1
        p2 = outport - 1

        H = self.Z[p2, p1] / self.Z[p1, p1]
        return LaplaceDomainVoltage(self.Voc[p2] + (V - self.Voc[p1]) * H)

    def Iresponse(self, I, inport=1, outport=2):
        """Return current response for specified current voltage and
        specified ports"""

        self._portcheck(inport)
        self._portcheck(outport)

        p1 = inport - 1
        p2 = outport - 1

        Y = self.Y
        Isc = self.Isc

        return LaplaceDomainCurrent(Isc[p2] + (I - Isc[p1]) * Y[p2, p1] / Y[p1, p1])

    def attach_parallel(self, OP, port=2):
        """Attach one-port in parallel to specified port"""

        if not issubclass(OP.__class__, OnePort):
            raise TypeError('Argument not ', OnePort)

        self._portcheck(port)

        p = port - 1

        Y = self.Y
        Y[p, p] += OP.Ys
        Isc = self.Isc
        Isc[p] += OP.Isc
        Z = Y.Z
        Voc = LaplaceDomainVoltageMatrix(
            [LaplaceDomainVoltage(Isc[m] * Z[m, m]) for m in range(len(Isc))])
        return ThreePort(Z, Voc)

    def bridge(self, OP, inport=1, outport=2):
        """Bridge the specified ports with a one-port element"""

        self._portcheck(inport)
        self._portcheck(outport)

        # Create two-port series element.
        s = Series(OP)

        # The impedance matrix for a series element is infinite.

        Y3 = YMatrix3(((0, 0, 0), (0, 0, 0), (0, 0, 0)))
        Y2 = s.Y
        p1 = inport - 1
        p2 = outport - 1

        Y3[p1, p1] = Y2[0, 0]
        Y3[p2, p2] = Y2[1, 1]
        Y3[p1, p2] = Y2[0, 1]
        Y3[p2, p1] = Y2[1, 0]

        Y = self.Y + Y3
        Isc = self.Isc
        Isc[p1] -= OP.Isc
        Isc[p2] += OP.Isc
        Z = Y.Z
        Voc = LaplaceDomainVoltageMatrix(
            [LaplaceDomainVoltage(Isc[m] * Z[m, m]) for m in range(len(Isc))])
        return ThreePort(Y.Z, Voc)

    def parallel(self, MP, port=None):
        """Return the model with, MP, in parallel"""

        if issubclass(MP.__class__, OnePort):
            return self.attach_parallel(MP, port)

        if issubclass(MP.__class__, TwoPort):
            # We could special case a series or shunt network here.
            raise NotImplementedError('TODO')

        if not issubclass(MP.__class__, ThreePort):
            raise TypeError('Argument not ', ThreePort)

        Y = self.Y + MP.Y
        Isc = self.Isc + MP.Isc
        Z = Y.Z
        Voc = LaplaceDomainVoltageMatrix(
            [LaplaceDomainVoltage(Isc[m] * Z[m, m]) for m in range(len(Isc))])
        return ThreePort(Z, Voc)

    def series(self, MP, port=None):
        """Return the model with, MP, in series"""

        if issubclass(MP.__class__, OnePort):
            raise NotImplementedError('TODO')

        if issubclass(MP.__class__, TwoPort):
            raise NotImplementedError('TODO')

        if not issubclass(MP.__class__, ThreePort):
            raise TypeError('Argument not ', ThreePort)

        warn('Will this ever work?')

        Z = self.Z + MP.Z
        Voc = self.Voc + MP.Voc

        return ThreePort(Z, Voc)

    def terminate(self, OP, port=2):
        """Connect one-port in parallel to specified port and return a
        two-port object"""

        return self.attach_parallel(OP, port).open_circuit(port)

    def short_circuit(self, port=2):
        """Apply a short-circuit to specified port and return a
        two-port object"""

        # Remove the unwanted port from the Ymatrix.
        Y = self.Y.copy()
        Y.row_del(port - 1)
        Y.col_del(port - 1)
        Y = YMatrix(Y)

        # CHECKME, perhaps use Isc?
        Voc = self.Voc.copy()
        Voc.row_del(port - 1)

        return TwoPortZModel(Y.Z, LaplaceDomainVoltage(Voc[0]), LaplaceDomainVoltage(Voc[1]))

    def open_circuit(self, port=2):
        """Apply a open-circuit to specified port and return a
        two-port object"""

        # Remove the unwanted port from the Zmatrix.
        Z = self.Z.copy()
        Z.row_del(port - 1)
        Z.col_del(port - 1)
        Z = ZMatrix(Z)

        Voc = self.Voc.copy()
        Voc.row_del(port - 1)

        return TwoPortZModel(Z, LaplaceDomainVoltage(Voc[0]), LaplaceDomainVoltage(Voc[1]))


class Opamp(ThreePort):

    """
    Create an ideal(ish) opamp
    ::

            |\
            |  \
        1 --+ +  \
            |      \
            |        \
            |          +--- 3
            |        /
            |      /
        2 --+ -  /
            |  /
            |/

        Each port voltage, Vn, is referenced to a common ground.
        Port 1: non-inverting input
        Port 2: inverting input
        Port 3: output

    """

    def __init__(self, Rd=1e9, Ro=1e-6, A=100000, Rp=1e9, Rm=1e9):

        # If Ro=0, then Z matrix singular.

        Rd, Ro, A, Rp, Rm = [cexpr(arg) for arg in (Rd, Ro, A, Rp, Rm)]

        Ra = Rp * (Rd + Rm) / (Rp + Rd + Rm)
        Rb = Rm * (Rd + Rp) / (Rp + Rd + Rm)

        Z = ZMatrix3(((Rp + Rd, Rd, 0),
                      (Rd, Rm + Rd, 0),
                      (A * Ra, -A * Rb, Ro)))
        super(Opamp, self).__init__(Z)
        self.args = (Rd, Ro, A, Rp, Rm)



"""
This module implements experimental network synthesis.

Copyright 2020 Michael Hayes, UCECE
"""

from .oneport import L, C, R, G, parallel, series
from .sexpr import s, LaplaceDomainExpression
from .exprclasses import LaplaceDomainImpedance
from .expr import expr

# Should check that args to L, C, R, G contains s and raise
# exception since the circuit is not realisable.

# Could make this a mixin for Zs and Ys but would need a Z flavour
# and a Y flavour.


class Synthesis(object):

    def seriesRL(self, lexpr):
        """Z = s * L + R"""

        if lexpr == 0:
            # Perhaps return Wire object
            return None

        var = lexpr.var
        lexpr = lexpr.partfrac()
        d = lexpr.expr.collect(var, evaluate=False)

        net = None
        a = d.pop(1, None)
        if a is not None:
            net = R(a)

        a = d.pop(var, None)
        if a is not None:
            net = series(net, L(a))

        if d != {}:
            raise ValueError('Not series RL')

        return net

    def seriesRC(self, lexpr):
        """Z = R + 1 / (s * C)"""

        if lexpr == 0:
            raise ValueError('Not series RC')

        var = lexpr.var
        lexpr = lexpr.partfrac()
        d = lexpr.expr.collect(var, evaluate=False)

        net = None
        a = d.pop(1, None)
        if a is not None:
            net = R(a)

        a = d.pop(1 / var, None)
        if a is not None:
            net = series(net, C(1 / a))

        if d != {}:
            raise ValueError('Not series RC')

        return net

    def seriesGC(self, lexpr):
        """Z = 1 / G + 1 / (s * C)"""

        if lexpr == 0:
            raise ValueError('Not series GC')

        var = lexpr.var
        lexpr = lexpr.partfrac()
        d = lexpr.expr.collect(var, evaluate=False)

        net = None
        a = d.pop(1, None)
        if a is not None:
            net = G(1 / a)

        a = d.pop(1 / var, None)
        if a is not None:
            net = series(net, C(1 / a))

        if d != {}:
            raise ValueError('Not series GC')

        return net

    def seriesLC(self, lexpr):
        """Z = s * L + 1 / (s * C)"""

        if lexpr == 0:
            raise ValueError('Not series LC')

        var = lexpr.var
        lexpr = lexpr.partfrac()
        d = lexpr.expr.collect(var, evaluate=False)

        net = None
        a = d.pop(1, 0)
        if a != 0:
            raise ValueError('Not series LC')

        a = d.pop(1 / var, None)
        if a is not None:
            net = C(1 / a)

        a = d.pop(var, None)
        if a is not None:
            net = series(net, L(a))

        if d != {}:
            raise ValueError('Not series LC')

        return net

    def seriesRLC(self, lexpr):
        """Z = s * L + R + 1 / (s * C)"""

        if lexpr == 0:
            raise ValueError('Not series RLC')

        var = lexpr.var
        lexpr = lexpr.partfrac()
        d = lexpr.expr.collect(var, evaluate=False)

        if len(d) > 3:
            raise ValueError('Not series RLC, degree > 2')

        net = None
        a = d.pop(1, None)
        if a is not None:
            net = R(a)

        a = d.pop(1 / var, None)
        if a is not None:
            net = series(net, C(1 / a))

        a = d.pop(var, None)
        if a is not None:
            net = series(net, L(a))

        if d != {}:
            raise ValueError('Not series RLC')

        return net

    def parallelRL(self, lexpr):
        """Y = 1 / (s * L) + 1 / R"""

        if lexpr == 0:
            # Perhaps return Wire object
            return None

        var = lexpr.var
        yexpr = (1 / lexpr).partfrac()
        d = yexpr.expr.collect(var, evaluate=False)

        net = None
        a = d.pop(1, None)
        if a is not None:
            net = R(1 / a)

        a = d.pop(1 / var, None)
        if a is not None:
            net = parallel(net, L(1 / a))

        if d != {}:
            raise ValueError('Not parallel RL')

        return net

    def parallelRC(self, lexpr):
        """Y = s * C + 1 / R"""

        if lexpr == 0:
            raise ValueError('Not parallel RC')

        var = lexpr.var
        yexpr = (1 / lexpr).partfrac()
        d = yexpr.expr.collect(var, evaluate=False)

        net = None
        a = d.pop(1, None)
        if a is not None:
            net = R(1 / a)

        a = d.pop(var, None)
        if a is not None:
            net = parallel(net, C(a))

        if d != {}:
            raise ValueError('Not parallel RC')

        return net

    def parallelGC(self, lexpr):
        """Y = s * C + G"""

        if lexpr == 0:
            raise ValueError('Not parallel GC')

        var = lexpr.var
        yexpr = (1 / lexpr).partfrac()
        d = yexpr.expr.collect(var, evaluate=False)

        net = None
        a = d.pop(1, None)
        if a is not None:
            net = G(a)

        a = d.pop(var, None)
        if a is not None:
            net = parallel(net, C(a))

        if d != {}:
            raise ValueError('Not parallel GC')

        return net

    def parallelLC(self, lexpr):
        """Y = s * C + 1 / (s * L)"""

        if lexpr == 0:
            raise ValueError('Not parallel LC')

        var = lexpr.var
        yexpr = (1 / lexpr).partfrac()
        d = yexpr.expr.collect(var, evaluate=False)

        net = None
        a = d.pop(1, 0)
        if a != 0:
            raise ValueError('Not parallel LC')

        a = d.pop(var, None)
        if a is not None:
            net = C(a)

        a = d.pop(1 / var, None)
        if a is not None:
            net = parallel(net, L(1 / a))

        if d != {}:
            raise ValueError('Not parallel LC')

        return net

    def parallelRLC(self, lexpr):
        """Y = s * C + 1 / R + 1 / (s * L)

        or Z = (s / C) / (s**2 + s / (C * R) + 1 / (C * L)).

        """

        if lexpr == 0:
            raise ValueError('Not parallel RLC')

        var = lexpr.var
        yexpr = (1 / lexpr).partfrac()
        d = yexpr.expr.collect(var, evaluate=False)

        if len(d) > 3:
            raise ValueError('Not parallel RLC, degree > 2')

        Rnet = None
        Lnet = None
        Cnet = None
        a = d.pop(1, None)
        if a is not None:
            Rnet = R(1 / a)

        a = d.pop(var, None)
        if a is not None:
            Cnet = C(a)

        a = d.pop(1 / var, None)
        if a is not None:
            Lnet = L(1 / a)

        if d != {}:
            raise ValueError('Not parallel RLC')

        return parallel(Rnet, Lnet, Cnet)

    def RLC(self, lexpr):
        """Handle series RLC or parallel RLC.  This will fail on R | L + C."""

        try:
            return self.seriesRLC(lexpr)
        except:
            return self.parallelRLC(lexpr)

    def parallelseriesRL(self):
        """(R('R1') + L('L1')) | (R('R2') + L('L2'))"""
        # Like Foster II

        raise NotImplementedError('TODO')

    def seriesparallelRL(self):
        """(R('R1') | L('L1')) + (R('R2') | L('L2'))"""
        # Like Foster I

        raise NotImplementedError('TODO')

    def fosterI(self, lexpr):

        expr = lexpr.partfrac(combine_conjugates=True)

        net = None
        for term in expr.as_ordered_terms():
            # Convert R - R**2/(L*(s + R/L))) to L*R*s/(L*s + R)
            # and handle mess from partfrac
            term = term.simplify()

            # This requires s / (s**2 + a*s + b)
            net = series(net, self.parallelRLC(LaplaceDomainExpression(term)))
        return net

    def fosterII(self, lexpr):

        expr = (1 / lexpr).partfrac(combine_conjugates=True)

        net = None
        for term in expr.as_ordered_terms():
            term = term.simplify()
            net = parallel(net, self.seriesRLC(
                LaplaceDomainExpression(1 / term)))
        return net

    def cauerI(self, lexpr):

        coeffs = lexpr.continued_fraction_coeffs()

        net = None
        for m, coeff in enumerate(reversed(coeffs)):
            n = len(coeffs) - m - 1
            if n & 1 == 0:
                net = series(net, self.seriesRL(
                    LaplaceDomainExpression(coeff)))
            else:
                net = parallel(net, self.parallelGC(
                    LaplaceDomainExpression(1 / coeff)))
        return net

    def cauerII(self, lexpr):

        coeffs = (1 / lexpr).continued_fraction_inverse_coeffs()

        net = None
        for m, coeff in enumerate(reversed(coeffs)):
            n = len(coeffs) - m - 1
            if n & 1 == 0:
                if not (n == 0 and coeff == 0):
                    net = parallel(net, self.seriesRL(
                        LaplaceDomainExpression(1 / coeff)))
            else:
                net = series(net, self.parallelGC(
                    LaplaceDomainExpression(coeff)))
        return net

    def network(self, lexpr, form='default'):

        if form == 'default':
            form = 'cauerI'

        lexpr = expr(lexpr)
        if not lexpr.is_impedance:
            raise ValueError('Expression needs to be an impedance')
        lexpr = lexpr.laplace()

        # Should test if a positive real function.

        try:
            method = getattr(self, form)
        except AttributeError:
            raise ValueError(
                '''Unknown form %s, known forms include: cauerI, cauerII, fosterI, fosterII''' % form)

        net = method(lexpr)
        return net


def network(lexpr, form='default'):
    """Synthesise a network with an equivalent impedance.  `form`
    includes: cauerI, cauerII, fosterI, fosterII, seriesRLC, seriesGC,
    parallelRLC, seriesRC, seriesRL, parallelRC, parallelRL,
    parallelGC.  The default is CauerI.

    Note some methods generate networks with negative value
    components."""

    return Synthesis().network(lexpr, form)
